import argparse
import os
import time

import numpy as np
import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

txaio.use_twisted()

ON_RTD = os.environ.get('READTHEDOCS') == 'True'
if not ON_RTD:
    import socs.agents.hwp_picoscope.drivers.class_ps3000a as ps


class PicoAgent:
    """Agent for interfacing with a Picoscope 3403D MSO device.

    Args:
        agent (ocs.ocs_agent.OCSAgent): Instantiated OCSAgent class for this
            Agent
    """

    def __init__(self, agent):

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.initialized = False
        self.take_data = False

        # Registers raw data and down sampled data feeds
        agg_params = {'frame_length': 60, 'exclude_influx': True}
        self.agent.register_feed('sensors', record=True, agg_params=agg_params)
        agg_params = {'frame_length': 60}
        self.agent.register_feed('downsampled_sensors', record=True, agg_params=agg_params)

    # Task functions.
    @ocs_agent.param('Npoints', default=10000, type=int)
    @ocs_agent.param('samplefreq', default=3.6e6, type=float)
    @ocs_agent.param('biasfreq', default=150e3, type=float)
    def run_single(self, session, params):
        """run_single(Npoints=10000, samplefreq=3.6e6, biasfreq=150e3)

           **Task** - Bias LC probes and perform DAQ.

           Parameters:
               Npoints (int): Number of points to measure
               samplefreq (float): sampling frequency (Hz), typically 24*biasfreq
               biasfrequency (float): LC probe bias frequency (Hz)

        """
        Npoints = int(params['Npoints'])
        samplefreq = float(params['samplefreq'])
        biasfreq = float(params['biasfreq'])

        with self.lock.acquire_timeout(0, job='run_single') as acquired:
            if not acquired:
                self.log.warn("Could not start run_single because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

        current_time = time.time()

        pico = ps.ps3000a(Npoints, samplefreq)
        pico.SigGenSingle(biasfreq)
        pico.SetScopeAll()
        pico.SetBufferAll()
        pico.set_digital_port()
        pico.set_digital_buffer()
        pico.Stream_AD()
        t, A, B, C, D = pico.get_value()
        Digital = pico.get_digital_values_simple()
        info = pico.info
        pico.close()

        # save downsampled data and metadata.
        data_downsampled = {
            'block_name': 'sens',
            'timestamp': current_time,
            'data': {}
        }
        for key, value in info.items():
            data_downsampled['data'][key] = value
        data_downsampled['data']['ch_A_ave'] = np.average(A)
        data_downsampled['data']['ch_B_ave'] = np.average(B)
        data_downsampled['data']['ch_C_ave'] = np.average(C)
        data_downsampled['data']['ch_D_ave'] = np.average(D)
        data_downsampled['data']['ch_A_std'] = np.std(A)
        data_downsampled['data']['ch_B_std'] = np.std(B)
        data_downsampled['data']['ch_C_std'] = np.std(C)
        data_downsampled['data']['ch_D_std'] = np.std(D)
        data_downsampled['data']['ch_B_Acos'] = np.average(np.array(A) * np.array(B))
        data_downsampled['data']['ch_C_Acos'] = np.average(np.array(A) * np.array(C))
        data_downsampled['data']['ch_D_Acos'] = np.average(np.array(A) * np.array(D))
        # approximated sin
        nroll = 6
        data_downsampled['data']['ch_B_Asin'] = np.average(np.roll(A, nroll) * np.array(B))
        data_downsampled['data']['ch_C_Asin'] = np.average(np.roll(A, nroll) * np.array(C))
        data_downsampled['data']['ch_D_Asin'] = np.average(np.roll(A, nroll) * np.array(D))
        for i, ch in enumerate(Digital):
            ch = np.array([int(d) for d in ch])
            rising_edge = list(np.flatnonzero((ch[:-1] < .5) & (ch[1:] > .5)) + 1)
            falling_edge = list(np.flatnonzero((ch[:-1] > .5) & (ch[1:] < .5)) + 1)
            # data_downsampled['data']['ch_%d_rise'%i] = rising_edge
            # data_downsampled['data']['ch_%d_fall'%i] = falling_edge
            data_downsampled['data']['ch_%d_rate' % i] = (len(rising_edge or []) + len(falling_edge or [])) / pico.info['length_sec']
        self.agent.publish_to_feed('downsampled_sensors', data_downsampled)
        print(data_downsampled['data'])

        # save raw data
        # split data and publish one by one
        Np = 10000
        for i in np.arange(int(np.ceil(len(t) / Np))):
            data = {
                'block_name': 'sens',
                'data': {}
            }
            t_split = list(t[i * Np:(i + 1) * Np])
            A_split = list(A[i * Np:(i + 1) * Np])
            B_split = list(B[i * Np:(i + 1) * Np])
            C_split = list(C[i * Np:(i + 1) * Np])
            D_split = list(D[i * Np:(i + 1) * Np])

            data['data']['timestamp'] = t_split
            data['data']['ch_A'] = A_split
            data['data']['ch_B'] = B_split
            data['data']['ch_C'] = C_split
            data['data']['ch_D'] = D_split
            data['timestamps'] = [current_time + i / samplefreq for i in range(len(t_split))]
            for j, ch in enumerate(Digital):
                ch = [int(d) for d in ch]
                data['data']['ch_%d' % j] = ch[i * Np:(i + 1) * Np]

            self.log.debug('publish {}/{}. {}'.format(i, int(np.ceil(len(t) / Np)), len(t_split)))
            self.agent.publish_to_feed('sensors', data)
            self.agent.feeds['sensors'].flush_buffer()

        return True, 'Single acquisition exited cleanly.'

    @ocs_agent.param('freq', default=10., type=float)
    @ocs_agent.param('duration', default=1., type=float)
    def sig_test(self, session, params):
        """sig_test(freq=10., duration=1.)

           **Task** - For debug and test.

           Parameters:
               freq (float): bias frequency (Hz)
               duration (float): bias duration time (sec)

        """
        with self.lock.acquire_timeout(0, job='sig_test') as acquired:
            if not acquired:
                self.log.warn("Could not start sig_test because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

        freq = params['freq']
        duration = params['duration']
        pico = ps.ps3000a(1, 1)
        pico.SigGenSingle(freq)
        time.sleep(duration)
        pico.close()
        return True, 'Signal generator test is done.'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    # Fix me after correcting "priviledged true"
    # pgroup = parser.add_argument_group('Agent Options')
    # pgroup.add_argument('--port', type=stri, help="Path to USB node for the picoscope.")
    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='PicoAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    pa = PicoAgent(agent)
    agent.register_task('run_single', pa.run_single)
    agent.register_task('sig_test', pa.sig_test)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import time
from os import environ

import txaio
from autobahn.twisted.util import sleep as dsleep
from ocs import ocs_agent, site_config
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

from socs.snmp import SNMPTwister

# For logging
txaio.use_twisted()


class MeinbergSNMP:
    """Meinberg SNMP communicator. Handles communication with and decoding of
    values from the Meinberg M1000.

    Parameters
    ----------
    address : str
        Address of the M1000.
    port : int
        SNMP port to issue GETs to, default to 161.
    version : int
        SNMP version for communication (1, 2, or 3), defaults to 3.

    Attributes
    ----------
    address : str
        Address of the M1000.
    port : int
        SNMP port to issue GETs to.
    snmp : socs.snmp.SNMPTwister
        snmp handler from SOCS
    mib_timings : list
        list of dicts describing the SNMP OIDs to check, and at which
        intervals. Each dict contains the keys "oid", "interval", and
        "lastGet". The corresponding values are of types tuple, integer, and
        float, respectively. "lastGet" is initialized as None, since no SNMP
        GET commands have been issued.
    oid_cache : dict
        Cache of OID values and corresponding decoded values. Meant to be
        passed directly to session.data.

    """

    def __init__(self, address, port=161, version=3):
        self.log = txaio.make_logger()
        self.address = address
        self.port = port
        self.version = version
        self.snmp = SNMPTwister(address, port)

        # OIDs and how often to query them
        self.mib_timings = [{"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgRefclockState', 1),
                             "interval": 60,
                             "lastGet": None},
                            {"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgSysPsStatus', 1),
                             "interval": 60,
                             "lastGet": None},
                            {"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgSysPsStatus', 2),
                             "interval": 60,
                             "lastGet": None},
                            {"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgEthPortLinkState', 1),
                             "interval": 60,
                             "lastGet": None},
                            {"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgRefclockLeapSecondDate', 1),
                             "interval": 60 * 60,
                             "lastGet": None},
                            {"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgNtpCurrentState', 0),
                             "interval": 64,
                             "lastGet": None},
                            {"oid": ('MBG-SNMP-LTNG-MIB', 'mbgLtNgPtpPortState', 1),
                             "interval": 3,
                             "lastGet": None}]

        self.oid_cache = {}

        # Determine unique interval values
        self.interval_groups = list({x['interval'] for x in self.mib_timings})

    def _build_get_list(self, interval):
        """Create list of OIDs to GET based on last time we checked them.

        Intervals are defined in the mib_timings dictionary. If the interval
        time, or more, has passed since the last GET was issued, we add the OID to the
        get_list.

        Parameters
        ----------
        interval : int
            Interval to build the get list for. Since available fields cannot
            change dynamically within OCS we publish each interval group to its
            own block.

        Returns
        -------
        get_list : list
            List of OID tuples to be passed to an SNMPTwister object in a GET call.

        """
        get_list = []
        for mib in self.mib_timings:
            # Select only OIDs of the same interval
            if mib['interval'] != interval:
                continue

            if mib["lastGet"] is None:
                get_list.append(mib["oid"])
            elif time.time() - mib["lastGet"] > mib["interval"]:
                get_list.append(mib["oid"])

        return get_list

    def _extract_oid_field_and_value(self, get_result):
        """Extract field names and OID values from SNMP GET results.

        The ObjectType objects returned from pysnmp interactions contain the
        info we want to use for field names, specifically the OID and associated
        integer for uniquely identifying duplicate OIDs, as well as the value of the
        OID, which we want to save.

        Here we use the prettyPrint() method to get the OID name, requiring
        some string manipulation. We also just grab the hidden ._value
        directly, as this seemed the cleanest way to get an actual value of a
        normal type. Without doing this we get a pysnmp defined Integer32 or
        DisplayString, which were akward to handle, particularly the
        DisplayString.

        Parameters
        ----------
        get_result : pysnmp.smi.rfc1902.ObjectType
            Result from a pysnmp GET command.

        Returns
        -------
        field_name : str
            Field name for an OID, i.e. 'mbgLtNgRefclockState_1'
        oid_value : int or str
            Associated value for the OID. Returns None if not an int or str
        oid_description : str
            String description of the OID value.

        """
        # OID from SNMP GET
        oid = get_result[0].prettyPrint()
        # Makes something like 'MBG-SNMP-LTNG-MIB::mbgLtNgRefclockState.1'
        # look like 'mbgLtNgRefclockState_1'
        field_name = oid.split("::")[1].replace('.', '_')

        # Grab OID value, mostly these are integers
        oid_value = get_result[1]._value
        oid_description = get_result[1].prettyPrint()

        self.log.debug("{o} {value}",
                       o=field_name,
                       value=oid_value)

        # Decode string values
        if isinstance(oid_value, bytes):
            oid_value = oid_value.decode("utf-8")

        # I don't expect any other types at the moment, but just in case.
        if not isinstance(oid_value, (int, bytes, str)):
            self.log.error("{oid} is of unknown and unhandled type "
                           + "{oid_type}. Returning None.",
                           oid=oid, oid_type=type(oid_value))
            oid_value = None

        return field_name, oid_value, oid_description

    def update_cache(self, get_result, timestamp):
        """Update the OID Value Cache.

        The OID Value Cache is used to store each unique OID, the latest value,
        the associated decoded string, and the last time the OID was queried from the
        M1000.

        The cache consists of a dictionary, with the unique OIDs as keys, and
        another dictionary as the value. Each of these nested dictionaries contains the
        OID values, description (decoded string), and last query time. An
        example for a single OID::

            {"mbgLtNgPtpPortState_1":
                {"status": 3,
                 "description": "disabled",
                 "lastGet": 1598543397.689727}}

        Additionally there is connection status information under::

            {"m1000_connection":
                {"last_attempt": 1598543359.6326838,
                "connected": True}

        This method modifies self.oid_cache.

        Parameters
        ----------
        get_result : pysnmp.smi.rfc1902.ObjectType
            Result from a pysnmp GET command.
        timestamp : float
            Timestamp for when the SNMP GET was issued.

        """
        try:
            for item in get_result:
                field_name, oid_value, oid_description = self._extract_oid_field_and_value(item)
                if oid_value is None:
                    continue

                # Update OID Cache for session.data
                self.oid_cache[field_name] = {"status": oid_value}
                self.oid_cache[field_name]["lastGet"] = timestamp
                self.oid_cache[field_name]["description"] = oid_description
                self.oid_cache['m1000_connection'] = {'last_attempt': time.time(),
                                                      'connected': True}
        # This is a TypeError due to nothing coming back from the yield in
        # run_snmp_get, so get_result is None here and can't be iterated.
        except TypeError:
            self.oid_cache['m1000_connection'] = {'last_attempt': time.time(),
                                                  'connected': False}
            raise ConnectionError('No SNMP response. Check your connection.')

    def _build_message(self, interval, get_result, time):
        """Build the message for publication on an OCS Feed.

        For a given MIB timing interval, build a message for Feed publication.
        Each interval contains only the OIDs that are sampled on the same timing
        interval. We split by interval since the available fields cannot change
        within a block over time.

        Parameters
        ----------
        interval : int
            Timing interval in seconds
        get_result : pysnmp.smi.rfc1902.ObjectType
            Result from a pysnmp GET command.
        time : float
            Timestamp for when the SNMP GET was issued.

        Returns
        -------
        message : dict
            OCS Feed formatted message for publishing

        """
        message = {
            'block_name': f'm1000_{interval}',
            'timestamp': time,
            'data': {}
        }

        for item in get_result:
            field_name, oid_value, oid_description = self._extract_oid_field_and_value(item)

            if oid_value is None:
                continue

            message['data'][field_name] = oid_value
            message['data'][field_name + "_description"] = oid_description

        return message

    @inlineCallbacks
    def run_snmp_get(self, session):
        """Peform the main data acquisition steps, issuing SNMP GET commands
        for each OID, depending on when we last queried them.

        These steps are performed for each group of OIDs in the same timing
        interval. We first build a list of OIDs to query. We then query them, update
        the local cache which is passed to the session.data object, build an OCS
        formatted message, and publish that message on the OCS Feed.

        If no OID should be queried yet we continue, expecting the Agent to
        handle any waiting that should be done between queries.

        """
        for interval in self.interval_groups:
            # Create list of OIDs to GET based on last time we checked them
            get_list = self._build_get_list(interval)

            # empty if an interval of time hasn't passed since last GET
            if not get_list:
                continue

            # Issue SNMP GET command
            result = yield self.snmp.get(get_list, self.version)
            read_time = time.time()

            # Do not publish if M1000 connection has dropped
            try:
                self.update_cache(result, read_time)
                message = self._build_message(interval, result, read_time)

                # Update lastGet time
                for mib in self.mib_timings:
                    if mib['oid'] in get_list:
                        mib['lastGet'] = read_time

                self.log.debug("{msg}", msg=message)
                session.app.publish_to_feed('m1000', message)
            except ConnectionError as e:
                self.log.error(f'{e}')

            # Update connection status in session.data
            session.data = self.oid_cache


class MeinbergM1000Agent:
    """Monitor the Meinberg LANTIME M1000 timing system via SNMP.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    address : str
        Address of the M1000.
    port : int
        SNMP port to issue GETs to, default to 161.
    version : int
        SNMP version for communication (1, 2, or 3), defaults to 3.

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    is_streaming : bool
        Tracks whether or not the agent is actively issuing SNMP GET commands
        to the M1000. Setting to false stops sending commands.
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent

    """

    def __init__(self, agent, address, port=161, version=3):
        self.agent = agent
        self.is_streaming = False
        self.log = self.agent.log

        self.log.info(f'Using SNMP version {version}.')
        self.meinberg = MeinbergSNMP(address, port, version)

        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('m1000',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @inlineCallbacks
    def acq(self, session, params=None):
        """acq()

        **Process** - Fetch values from the M1000 via SNMP.

        Notes:
            The session.data object stores each unique OID with its latest status,
            decoded value, and the last time the value was retrived. This will look like
            the example here::

                >>> response.session['data']
                {"mbgLtNgRefclockLeapSecondDate_1":
                    {"status": "not announced",
                     "lastGet":1598626144.5365012},
                 "mbgLtNgPtpPortState_1":
                    {"status": 3,
                     "description": "disabled",
                     "lastGet": 1598543397.689727},
                 "mbgLtNgNtpCurrentState_0":
                    {"status": 1,
                     "description": "not synchronized",
                     "lastGet": 1598543363.289597},
                 "mbgLtNgRefclockState_1":
                    {"status": 2,
                     "description": "not synchronized",
                     "lastGet": 1598543359.6326838},
                 "mbgLtNgSysPsStatus_1":
                    {"status": 2,
                     "description": "up",
                     "lastGet": 1598543359.6326838},
                 "mbgLtNgSysPsStatus_2":
                    {"status": 2,
                     "description": "up",
                     "lastGet": 1598543359.6326838},
                 "mbgLtNgEthPortLinkState_1":
                    {"status": 1,
                     "description": "up",
                     "lastGet": 1598543359.6326838}
                 "m1000_connection":
                    {"last_attempt": 1598543359.6326838,
                     "connected": True}}

            Note that session.data is populated within the :func:`MeinbergSNMP.run_snmp_get` call.

        """
        if params is None:
            params = {}

        # Make an initial attempt at connection.
        # Allows us to fail early if misconfigured.
        yield self.meinberg.run_snmp_get(session)
        if not self.meinberg.oid_cache['m1000_connection'].get('connected', False):
            self.log.error('No initial SNMP response.')
            self.log.error('Either there is a network connection issue, '
                           + 'or maybe you are using the wrong SNMP '
                           + 'version. Either way, we are exiting.')

            reactor.callFromThread(reactor.stop)
            return False, 'acq process failed - No connection to M1000'

        self.is_streaming = True

        while self.is_streaming:
            yield self.meinberg.run_snmp_get(session)
            self.log.debug("{data}", data=session.data)
            yield dsleep(1)

        return True, "Finished Recording"

    def _stop_acq(self, session, params=None):
        """_stop_acq()

        **Task** - Stop task associated with acq process.

        """
        self.is_streaming = False
        return True, "Stopping Recording"


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--auto-start", default=True, type=bool,
                        help="Automatically start polling for data at "
                        + "Agent startup.")
    pgroup.add_argument("--address", help="Address to listen to.")
    pgroup.add_argument("--port", default=161,
                        help="Port to listen on.")
    pgroup.add_argument("--snmp-version", default='3', choices=['1', '2', '3'],
                        help="SNMP version for communication. Must match "
                             + "configuration on the M1000.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class="MeinbergM1000Agent",
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    listener = MeinbergM1000Agent(agent,
                                  address=args.address,
                                  port=int(args.port),
                                  version=int(args.snmp_version))

    agent.register_process("acq",
                           listener.acq,
                           listener._stop_acq,
                           startup=bool(args.auto_start), blocking=False)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



import time

import numpy as np


def constant_velocity_scanpoints(azpts, el, azvel, acc, ntimes):
    """
    Produces lists of times, azimuths, elevations, azimuthal velocities,
    elevation velocities, azimuth motion flags, and elevation motion flags
    for a finitely long azimuth scan with constant velocity. Scan begins
    at the first azpts value.

    Parameters:
        azpts (2-tuple): The endpoints of motion in azimuth, where the
            first point is the start position of the scan.
        el (float): The elevation that is maintained throughout the scan
        azvel (float): Desired speed of the azimuth motion in degrees/sec
        acc (float): The turnaround acceleration in degrees/sec^2
        ntimes(int): Number of times to travel between the endpoints.
            ntimes = 1 corresponds to a scan from, ex., left to right, and does not
            return to left.

    Returns:
        tuple of lists : (times, azimuths, elevations, azimuth veolicities,
        elevation velocities, azimuth flags, elevation flags)
    """
    if float(azvel) == 0.0:
        print('Azimuth velocity is zero, invalid scan parameter')
        return False
    if float(acc) == 0.0:
        print('Acceleration is zero, unable to calculate turnaround')
        return False
    turn_time = 2 * azvel / acc
    tot_time_dir = float((abs(azpts[1] - azpts[0])) / azvel)
    num_dirpoints = int(tot_time_dir * 10.)
    if num_dirpoints < 2:
        print('Scan is too short to run')
        return False
    sect_start_time = 0.0
    conctimes = []
    concaz = []
    el1 = np.linspace(el, el, num_dirpoints)
    concel = list(el1)
    concva = []
    ve1 = np.zeros(num_dirpoints)
    concve = list(ve1)

    # Flag values:
    # 0 : unidentified portion of the scan
    # 1 : constant velocity, with the next point at the same velocity
    # 2 : final point before a turnaround
    azflags = [1 for i in range(num_dirpoints - 1)]
    azflags += [2]
    all_azflags = []

    elflags = [0 for i in range(num_dirpoints)]
    all_elflags = []

    for n in range(ntimes):
        # print(str(n)+' '+str(sect_start_time))
        end_dir_time = sect_start_time + tot_time_dir
        time_for_section = np.linspace(sect_start_time, end_dir_time,
                                       num_dirpoints)
        if n % 2 != 0:
            new_az = np.linspace(azpts[1], azpts[0], num_dirpoints)
            new_va = np.zeros(num_dirpoints) + \
                np.sign(azpts[0] - azpts[1]) * azvel
        else:
            new_az = np.linspace(azpts[0], azpts[1], num_dirpoints)
            new_va = np.zeros(num_dirpoints) + \
                np.sign(azpts[1] - azpts[0]) * azvel

        conctimes.extend(time_for_section)
        concaz.extend(new_az)
        concel.extend(el1)
        concva.extend(new_va)
        concve.extend(ve1)
        all_azflags.extend(azflags)
        all_elflags.extend(elflags)
        sect_start_time = time_for_section[-1] + turn_time

    all_azflags[-1] = 0

    return conctimes, concaz, concel, concva, concve, all_azflags, \
        all_elflags


def from_file(filename):
    """
    Produces properly formatted lists of times, azimuth and elevation
    locations, azimuth and elevation velocities, and azimuth and elevation
    motion flags for a finitely long scan from a numpy file. Numpy file
    must be formatted as an array of arrays in the order [times, azimuths,
    elevations, azimuth velocities, elevation velocities]

    Parameters:
        filename (str): Full path to the numpy file containing scan
            parameter array

    Returns:
        tuple of lists: (times, azimuths, elevations, azimuth velocities,
        elevation velocities, azimuth flags, elevation flags)

    NOTE: Flags can be set in the numpy file (0=unspecified, 1=constant
    velocity, 2=last point before turnaround). If flags are not set in
    the file, all flags are set to 0 to accommodate non-linear scans
    """
    info = np.load(filename)
    if len(info) < 5:
        raise TypeError('Not enough fields in numpy file! Expected '
                        '5 fields.')
    conctimes = info[0]
    concaz = info[1]
    concel = info[2]
    concva = info[3]
    concve = info[4]
    if len(info) == 5:
        az_flags = np.array([0 for x in range(len(conctimes))])
        el_flags = az_flags
    elif len(info) == 7:
        az_flags = info[5]
        el_flags = info[6]
    else:
        print('File has too many parameters!')
        return False
    return conctimes, concaz, concel, concva, concve, az_flags, el_flags


def ptstack_format(conctimes, concaz, concel, concva, concve, az_flags,
                   el_flags, start_offset=3., generator=False):
    """
    Produces a list of lines in the format necessary to upload to the ACU
    to complete a scan. Params are the outputs of from_file,
    constant_velocity_scanpoints, or generate_constant_velocity_scan.

    Parameters:
        conctimes (list): Times starting at 0 for the ACU to reach
            associated positions
        concaz (list): Azimuth positions associated with conctimes
        concel (list): Elevation positions associated with conctimes
        concva (list): Azimuth velocities associated with conctimes
        concve (list): Elevation velocities associated with conctimes
        az_flags (list): Flags associated with azimuth motions at
            conctimes
        el_flags (list): Flags associated with elevation motions at
            conctimes
        start_offset (float): Seconds to wait before starting the scan
        generator (bool): Toggles the start time. When true, start time is
            start_offset, otherwise start time is time.time() + start_offset

    Returns:
        list: Lines in the correct format to upload to the ACU
    """

    fmt = '%j, %H:%M:%S'
    if generator:
        start_time = start_offset
    else:
        start_time = time.time() + start_offset
    true_times = [start_time + i for i in conctimes]
    fmt_times = [time.strftime(fmt, time.gmtime(t))
                 + ('{tt:.6f}'.format(tt=t % 1.))[1:] for t in true_times]

    all_lines = [('{ftime}; {az:.6f}; {el:.6f}; {azvel:.4f}; '
                  '{elvel:.4f}; {azflag}; {elflag}'
                  '\r\n'.format(ftime=fmt_times[n], az=concaz[n],
                                el=concel[n], azvel=concva[n], elvel=concve[n],
                                azflag=az_flags[n], elflag=el_flags[n]))
                 for n in range(len(fmt_times))]

    return all_lines


def generate_constant_velocity_scan(az_endpoint1, az_endpoint2, az_speed,
                                    acc, el_endpoint1, el_endpoint2,
                                    el_speed, num_batches=None,
                                    start_time=None, wait_to_start=3.,
                                    step_time=0.1, batch_size=500,
                                    az_start='mid_inc', ptstack_fmt=True):
    """
    Python generator to produce times, azimuth and elevation positions,
    azimuth and elevation velocities, azimuth and elevation flags for
    arbitrarily long constant-velocity azimuth scans.

    Parameters:
        az_endpoint1 (float): azimuth endpoint for the scan start
        az_endpoint2 (float): second azimuth endpoint of the scan
        az_speed (float): speed of the constant-velocity azimuth motion
        acc (float): turnaround acceleration for the azimuth motion at the
            endpoints
        el_endpoint1 (float): elevation endpoint for the scan start
        el_endpoint2 (float): second elevation endpoint of the scan. For
            constant az scans, this must be equal to el_endpoint1.
        el_speed (float): speed of the elevation motion. For constant az
            scans, set to 0.0
        num_batches (int or None): sets the number of batches for the
            generator to create. Default value is None (interpreted as infinite
            batches).
        start_time (float or None): a ctime at which to start the scan.
            Default is None, which is interpreted as starting now +
            wait_to_start.
        wait_to_start (float): number of seconds to wait between
            start_time and when the scan actually starts. Default is 3 seconds.
        step_time (float): time between points on the constant-velocity
            parts of the motion. Default value is 0.1 seconds. Minimum value is
            0.05 seconds.
        batch_size (int): number of values to produce in each iteration.
            Default is 500. Batch size is reset to the length of one leg of the
            motion if num_batches is not None.
        az_start (str): part of the scan to start at. Options are:
            'az_endpoint1', 'az_endpoint2', 'mid_inc' (start in the middle of
            the scan and start with increasing azimuth), 'mid_dec' (start in
            the middle of the scan and start with decreasing azimuth).
        ptstack_fmt (bool): determine whether values are produced with the
            necessary format to upload to the ACU. If False, this function will
            produce lists of time, azimuth, elevation, azimuth velocity,
            elevation velocity, azimuth flags, and elevation flags. Default is
            True.
    """
    az_min = min(az_endpoint1, az_endpoint2)
    az_max = max(az_endpoint1, az_endpoint2)
    if az_max == az_min:
        raise ValueError('Generator requires two different az endpoints!')
    if az_start in ['az_endpoint1', 'az_endpoint2']:
        if az_start == 'az_endpoint1':
            az = az_endpoint1
        else:
            az = az_endpoint2
        if az == az_min:
            increasing = True
            az_vel = az_speed
        elif az == az_max:
            increasing = False
            az_vel = -1 * az_speed
    elif az_start in ['mid_inc', 'mid_dec']:
        az = (az_endpoint1 + az_endpoint2) / 2
        if az_start == 'mid_inc':
            increasing = True
            az_vel = az_speed
        else:
            increasing = False
            az_vel = -1 * az_speed
    else:
        raise ValueError('az_start value not supported. Choose from '
                         'az_endpoint1, az_endpoint2, mid_inc, mid_dec')
    if start_time is None:
        t0 = time.time() + wait_to_start
    else:
        t0 = start_time
    t = 0
    turntime = 2.0 * az_speed / acc
    el = el_endpoint1
    if step_time < 0.05:
        raise ValueError('Time step size too small, must be at least '
                         '0.05 seconds')
    daz = step_time * az_speed
    el_vel = el_speed
    az_flag = 0
    el_flag = 0
    if num_batches is None:
        stop_iter = float('inf')
    else:
        stop_iter = num_batches
        batch_size = int(np.ceil((az_max - az_min) / daz))
    i = 0
    while i < stop_iter:
        i += 1
        point_block = [[], [], [], [], [], [], []]
        for j in range(batch_size):
            point_block[0].append(t + t0)
            point_block[1].append(az)
            point_block[2].append(el)
            point_block[3].append(az_vel)
            point_block[4].append(el_vel)
            point_block[5].append(az_flag)
            point_block[6].append(el_flag)
            t += step_time
            if increasing:
                if az <= (az_max - 2 * daz):
                    az += daz
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = True
                elif az == az_max:
                    t += turntime
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = False
                else:
                    az_remaining = az_max - az
                    time_remaining = az_remaining / az_speed
                    az = az_max
                    t += (time_remaining - step_time)
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 2
                    el_flag = 0
                    increasing = True
            else:
                if az >= (az_min + 2 * daz):
                    az -= daz
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = False
                elif az == az_min:
                    t += turntime
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = True
                else:
                    az_remaining = az - az_min
                    time_remaining = az_remaining / az_speed
                    az = az_min
                    t += (time_remaining - step_time)
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 2
                    el_flag = 0
                    increasing = False
        if ptstack_fmt:
            yield ptstack_format(point_block[0], point_block[1],
                                 point_block[2], point_block[3],
                                 point_block[4], point_block[5],
                                 point_block[6], generator=True)
        else:
            yield (point_block[0], point_block[1], point_block[2],
                   point_block[3], point_block[4], point_block[5],
                   point_block[6])


# if __name__ == "__main__":
#    print(time.time())
#    times, azs, els, vas, ves, azf, elf = linear_turnaround_scanpoints((120., 130.), 55., 1., 4, 2000)
#    print(time.time())



import argparse
import os
import time

import txaio
import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

ON_RTD = os.environ.get("READTHEDOCS") == "True"
if not ON_RTD:
    from holog_daq import synth3


class SynthAgent:
    """
    Agent for connecting to the Synths for holography.

    Args:
        config_file (str): ocs-site-configs/uchicago/field/holog_config.yaml
    """

    def __init__(self, agent, config_file):

        self.lo_id = None
        self.initialized = False
        self.take_data = False

        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        agg_params = {"frame_length": 10 * 60}  # [sec]
        self.agent.register_feed(
            "synth_lo", record=True, agg_params=agg_params, buffer_time=0
        )

        if config_file is None:
            raise Exception("No config file specified for the FTS mirror config")
        else:
            config_file_path = os.path.join(os.environ["OCS_CONFIG_DIR"], config_file)

            with open(config_file_path) as stream:
                self.holog_configs = yaml.safe_load(stream)
                if self.holog_configs is None:
                    raise Exception("No mirror configs in config file.")
                self.log.info(f"Loaded mirror configs from file {config_file_path}")
                self.N_MULT = self.holog_configs.pop("N_MULT", None)
                self.ghz_to_mhz = self.holog_configs.pop("ghz_to_mhz", None)

    def init_synth(self, session, params=None):
        """init_synth()

        **Task** - A task to initialize the synthesizers.

        Examples:
            Example for calling in a client::

                from ocs.ocs_client import OCSClient
                client = OCSClient("synth_lo")
                client.init_synth()

        Notes:
            This task is called to turn on the synthesizers.
        """

        self.log.debug("Trying to acquire lock")
        with self.lock.acquire_timeout(timeout=0, job="init") as acquired:
            # Locking mechanism stops code from proceeding if no lock acquired
            if not acquired:
                self.log.warn(
                    "Could not start init because {} is already running".format(
                        self.lock.job
                    )
                )
                return False, "Could not acquire lock."
            # Run the function you want to run
            self.log.debug("Lock Acquired Connecting to Stages")
            self.lo_id = synth3.get_LOs()

            synth3.set_RF_output(0, 1, self.lo_id)  # LO ID, On=1, USB connection ID
            synth3.set_RF_output(1, 1, self.lo_id)  # LO ID, On=1, USB connection ID

            # data = {"timestamp": time.time(), "block_name": "synth_lo", "data": {}}

            # data["data"]["F1_status"] = 1
            # data["data"]["F2_status"] = 1

            # self.agent.publish_to_feed("synth_lo", data)
            # session.data.update(data["data"])

        # This part is for the record and to allow future calls to proceed,
        # so does not require the lock
        self.initialized = True

        return True, "Synth Initialized."

    @ocs_agent.param("offset", type=float, default=0, check=lambda x: 0 <= x <= 1000)
    @ocs_agent.param("freq1", type=float, default=0, check=lambda x: 0 <= x <= 1000)
    def set_frequencies(self, session, params):
        """set_frequencies(freq1=0, offset=0)

        **Task** - A task to set the frequencies of the synthesizers.

        Parameters:
            freq1 (float): Frequency of holography measuremnt [GHz].
            offset (float): Frequency offset of holography measurement [MHz].

        Examples:
            Example for calling in a client::

                from ocs.ocs_client import OCSClient
                client = OCSClient("synth_lo")
                client.set_frequencies(freq1=210, offset=10)

        Notes:
            An example of the session data::

                >>> response.session['data']
                {"timestamp": 1601924482.722671,
                 "block_name": "synth_lo",
                 "data": {"F1": 11 ,
                          "F2": 11.1}}}
        """
        f1 = params.get("freq1", 0)
        f_offset = params.get("offset", 0)

        F_1 = int(f1 * self.ghz_to_mhz / self.N_MULT)  # Convert GHz -> MHz for synthesizers

        with self.lock.acquire_timeout(timeout=3, job="set_frqeuencies") as acquired:
            if not acquired:
                self.log.warn(
                    f"Could not set position because lock held by {self.lock.job}"
                )
                return False, "Could not acquire lock"

            synth3.set_f(0, F_1, self.lo_id)
            synth3.set_f(1, F_1 + f_offset, self.lo_id)

            data = {"timestamp": time.time(), "block_name": "synth_lo", "data": {}}

            data["data"]["F1"] = F_1
            data["data"]["F2"] = F_1 + f_offset

            self.agent.publish_to_feed("synth_lo", data)
            session.data.update(data["data"])

        return True, "Frequencies Updated"

    # This function is not finished, need to figure out how to read out frequency from USB connection.
    # def read_frequencies(self, session, params=None):

    #     with self.lock.acquire_timeout(timeout=3, job="read_frqeuencies") as acquired:
    #         if not acquired:
    #             self.log.warn(
    #                 f"Could not set position because lock held by {self.lock.job}"
    #             )
    #             return False, "Could not acquire lock"

    #     return True, "Frequencies Updated"

    @ocs_agent.param("lo_id", type=int, default=0, choices=[0, 1])
    @ocs_agent.param("status", type=int, default=0, choices=[0, 1])
    def set_synth_status(self, session, params):
        """set_synth_status(lo_id=0, status=1)

        **Task** - A task to set the status of the synthesizers.

        Parameters:
            lo_id (int): Local Oscillator ID (either 0 or 1).
            status (int): Status of local oscillator (0 is off, 1 is on).

        Examples:
            Example for calling in a client::

                from ocs.ocs_client import OCSClient
                agent = OCSClient("synth_lo")
                agent.set_synth_status(lo_id=0, status=1)

        Notes:
            An example of the session data::

                >>> response.session['data']
                {"timestamp": 1601924482.722671,
                 "block_name": "synth_lo",
                 "data": {"F1_status": 1}}
        """
        lo_id = params.get("lo_id", 0)
        switch = params.get("switch", 0)

        with self.lock.acquire_timeout(timeout=3, job="turn_on_or_off_synth") as acquired:
            if not acquired:
                self.log.warn(
                    f"Could not set position because lock held by {self.lock.job}"
                )
                return False, "Could not acquire lock"

            synth3.set_RF_output(lo_id, switch, self.lo_id)

            data = {"timestamp": time.time(), "block_name": "synth_lo", "data": {}}

            data["data"]["F1_status"] = switch

            self.agent.publish_to_feed("synth_lo", data)
            session.data.update(data["data"])

        return True, "Frequencies Updated"


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--config_file")

    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class="SynthAgent",
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    synth_agent = SynthAgent(agent, args.config_file)

    agent.register_task("init_synth", synth_agent.init_synth)
    agent.register_task("set_frequencies", synth_agent.set_frequencies)
    # agent.register_task("read_frequencies", synth_agent.read_frequencies)
    agent.register_task("set_synth_status", synth_agent.set_synth_status)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



import argparse
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

import socs.agents.wiregrid_actuator.limitswitch_config as limitswitch_config
import socs.agents.wiregrid_actuator.stopper_config as stopper_config
from socs.agents.wiregrid_actuator.drivers.Actuator import Actuator


class WiregridActuatorAgent:
    """ Agent to control the linear actuator
    to insert or eject the wire-grid via a GALIL motor controller.
    It communicates with the controller via an ethernet.
    It also reads ON/OFF of the limit-switches on the ends of the actuators
    and lock/unlock the stoppers to lock/unlock the actuators.

    Args:
        ip_address      (str): IP address for the GALIL motor controller
        interval_time (float): Interval time for dat acquisition
        sleep         (float): sleep time for every commands
        the motor controller
    """

    def __init__(self, agent, ip_address='192.168.1.100',
                 interval_time=1., sleep=0.05, verbose=0):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.run_acq = False
        self.ip_address = ip_address
        self.interval_time = interval_time
        self.sleep = sleep
        self.max_check_stopper = 100
        self.verbose = verbose

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'wgactuator', record=True, agg_params=agg_params)

        try:
            self.actuator = Actuator(
                self.ip_address, sleep=self.sleep,
                ls_list=limitswitch_config.IO_INFO,
                st_list=stopper_config.IO_INFO,
                verbose=self.verbose)
        except Exception as e:
            msg = '__init__(): Failed to initialize Actuator instance! '\
                  '| Exception = "{}"'.format(e)
            self.log.error(msg)
            self.actuator = None
            raise e

    ######################
    # Internal functions #
    ######################
    # Return: status(True or False), message
    # If an error occurs, return False.

    # Return: True/False, message, limit-switch ON/OFF
    def _move(self, distance, speedrate, LSLname, LSRname, LSlabel):
        LSL = 0  # left  actuator limit-switch
        LSR = 0  # right actuator limit-switch
        LSL, LSR = \
            self.actuator.ls.get_onoff(io_name=[LSLname, LSRname])
        if LSL == 0 and LSR == 0:
            ret = self.actuator.move(distance, speedrate)
            if not ret:
                msg = '_move(): WARNING!: Failed to move.'
                self.log.warn(msg)
                return False, msg, LSL or LSR
        else:
            self.log.warn(
                '_move(): One of {} limit-switches is ON (LSL={}, LSR={})!'
                .format(LSlabel, LSL, LSR))
            self.log.warn('  --> Did not move.')
        isrun = True
        # Loop until the limit-switch is ON or the actuator moving finishes
        while LSL == 0 and LSR == 0 and isrun:
            LSL, LSR = \
                self.actuator.ls.get_onoff(io_name=[LSLname, LSRname])
            isrun = self.actuator.is_run()
            if self.verbose > 0:
                self.log.info(
                    '_move(): LSL={}, LSR={}, run={}'.format(LSL, LSR, isrun))
        # Stop the actuator moving
        self.actuator.stop()
        LSonoff = LSL or LSR
        if LSonoff:
            self.log.info(
                '_move(): Stopped moving because '
                'one of {} limit-switches is ON (LSL1={}, LSR1={})!'
                .format(LSlabel, LSL, LSR))
        return True, \
            '_move(): Finish move(distance={}, speedrate={}, limit-switch={})'\
            .format(distance, speedrate, LSonoff), \
            LSonoff

    # Return value: True/False, message, limit-switch ON/OFF
    def _forward(self, distance, speedrate=0.2):
        if distance < 0.:
            distance = abs(distance)
        ret, msg, LSonoff = self._move(
            distance, speedrate, 'LSL1', 'LSR1', 'inside')
        return ret, \
            '_forward(): '\
            'Finish forward(distance={}, speedrate={}, limit-switch={})'\
            .format(distance, speedrate, LSonoff), \
            LSonoff

    # Return value: True/False, message, limit-switch ON/OFF
    def _backward(self, distance, speedrate=0.2):
        if distance > 0.:
            distance = -1. * abs(distance)
        ret, msg, LSonoff = self._move(
            distance, speedrate, 'LSL2', 'LSR2', 'outside')
        return ret, \
            '_backward(): '\
            'Finish backward(distance={}, speedrate={}, limit-switch={})'\
            .format(distance, speedrate, LSonoff), \
            LSonoff

    def _insert_eject(
            self, main_distance=920, main_speedrate=1.0, is_insert=True):
        # Function label
        flabel = 'insert' if is_insert else 'eject'
        initial_ls_names = ['LSL2', 'LSR2'] if is_insert else ['LSL1', 'LSR1']
        move_func = self._forward if is_insert else self._backward

        # Check connection
        ret = self.actuator.check_connect()
        if ret:
            if self.verbose > 0:
                self.log.info(
                    '_insert_eject()[{}]: '
                    'the connection to the actuator controller is OK!'
                    .format(flabel))
        else:
            msg = '_insert_eject()[{}]: ERROR!: '\
                  'the connection to the actuator controller is BAD!'\
                  .format(flabel)
            self.log.error(msg)
            return False, msg

        # Release stopper (Powering ON the stoppers)
        # 1st trial
        try:
            self.actuator.st.set_allon()
        except Exception as e:
            msg = '_insert_eject()[{}]: '\
                  'ERROR!: Failed to run the stopper set_allon() '\
                  '--> Stop moving! | Exception = "{}"'.format(flabel, e)
            self.log.error(msg)
            return False, msg
        # Check the stopper
        onoff_st = self.actuator.st.get_onoff()
        if not all(onoff_st):
            # 2nd trial (double check)
            try:
                self.actuator.st.set_allon()
            except Exception as e:
                msg = '_insert_eject()[{}]: '\
                      'ERROR!: Failed to run the stopper set_allon() '\
                      '--> Stop moving! | Exception = "{}"'.format(flabel, e)
                self.log.error(msg)
                return False, msg
            onoff_st = self.actuator.st.get_onoff()
            # Error in powering ON stopper
            if not all(onoff_st):
                msg = '_insert_eject()[{}]: '\
                      'ERROR!: Could not confirm all the stopper released '\
                      'after twice stopper set_allon()! '\
                      '--> Stop moving!'.format(flabel)
                self.log.error(msg)
                return False, msg

        # Initial slow & small moving
        ret, msg, LSonoff = move_func(5, speedrate=0.2)
        # Check the status of the initial moving
        if not ret:
            msg = '_insert_eject()[{}]: ERROR!: in the initail moving | {} '\
                  '--> Stop moving!'.format(flabel, msg)
            self.log.error(msg)
            # Lock the actuator by the stoppers
            self.actuator.st.set_alloff()
            return False, msg
        if LSonoff:
            msg = '_insert_eject()[{}]: WARNING!: '\
                  'Limit-switch is ON after the initial moving. '\
                  '---> Stop moving!'.format(flabel)
            self.log.warn(msg)
            # Lock the actuator by the stoppers
            self.actuator.st.set_alloff()
            return True, msg
        # Check limit-switch
        LSL, LSR = \
            self.actuator.ls.get_onoff(io_name=initial_ls_names)
        if LSL == 1 or LSR == 1:
            msg = '_insert_eject()[{}]: ERROR!: '\
                  'The limit-switch is NOT OFF '\
                  'after the initial moving. '\
                  '(Maybe the limit-switch is disconnected?) '\
                  '--> Stop moving!'.format(flabel)
            self.log.error(msg)
            # Lock the actuator by the stoppers
            self.actuator.st.set_alloff()
            return False, msg

        # Sleep before the main forwarding
        time.sleep(1)

        # Main moving
        status, msg, LSonoff = \
            move_func(main_distance, speedrate=main_speedrate)
        if not status:
            msg = '_insert_eject()[{}]: ERROR!: '\
                  'in the main moving | {} '\
                  '--> Stop moving!'.format(flabel, msg)
            self.log.error(msg)
            return False, msg
        if LSonoff:
            msg = '_insert_eject()[{}]: WARNING!: '\
                  'Limit-switch is ON after the main moving. '\
                  '---> Stop moving!'.format(flabel)
            self.log.warn(msg)
            # Lock the actuator by the stoppers
            self.actuator.st.set_alloff()
            return True, msg

        # Last slow & small moving
        status, msg, LSonoff = move_func(200, speedrate=0.2)
        if not status:
            msg = '_insert_eject()[{}]: ERROR!: in the last moving | {}'\
                  .format(flabel, msg)
            self.log.error(msg)
            return False, msg
        if LSonoff == 0:
            msg = '_insert_eject()[{}]: ERROR!: '\
                  'The limit-switch is NOT ON after last moving.'\
                  .format(flabel)
            self.log.error(msg)
            return False, msg

        # Lock the actuator by the stoppers
        self.actuator.st.set_alloff()
        # Check the stopper until all the stoppers are OFF (locked)
        for i in range(self.max_check_stopper):
            onoff_st = self.actuator.st.get_onoff()
            if not any(onoff_st):
                break
        if any(onoff_st):
            msg = '_insert_eject()[{}]: ERROR!: (After the last moving) '\
                  'Failed to lock (OFF) all the stoppers'.format(flabel)
            self.log.error(msg)
            return False, msg

        return True, '_insert_eject()[{}]: Successfully moving!'.format(flabel)

    def _insert(self, main_distance=920, main_speedrate=1.0):
        ret, msg = self._insert_eject(
            main_distance=main_distance, main_speedrate=main_speedrate,
            is_insert=True)
        return ret, msg

    def _eject(self, main_distance=920, main_speedrate=1.0):
        if main_distance > 0.:
            main_distance = -1. * abs(main_distance)
        ret, msg = self._insert_eject(
            main_distance=main_distance, main_speedrate=main_speedrate,
            is_insert=False)
        return ret, msg

    ##################
    # Main functions #
    ##################
    # Return: status(True or False), message

    def insert(self, session, params=None):
        """insert()

        **Task** - Insert the wire-grid into the forebaffle interface above the
        SAT.

        """
        with self.lock.acquire_timeout(timeout=3, job='insert') as acquired:
            if not acquired:
                self.log.warn(
                    'insert(): '
                    'Lock could not be acquired because it is held by {}.'
                    .format(self.lock.job))
                return False, 'insert(): Could not acquire lock'
            # Wait for a second before moving
            time.sleep(1)
            # Moving commands
            ret, msg = self._insert(920, 1.0)
            if not ret:
                msg = 'insert(): '\
                      'ERROR!: Failed insert() in _insert(850,1.0) | {}'\
                      .format(msg)
                self.log.error(msg)
                return False, msg
            return True, 'insert(): Successfully finish!'

    def eject(self, session, params=None):
        """eject()

        **Task** - Eject the wire-grid from the forebaffle interface above the
        SAT.

        """
        with self.lock.acquire_timeout(timeout=3, job='eject') as acquired:
            if not acquired:
                self.log.warn(
                    'eject(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'eject(): Could not acquire lock'
            # Wait for a second before moving
            time.sleep(1)
            # Moving commands
            ret, msg = self._eject(920, 1.0)
            if not ret:
                msg = 'eject(): ERROR!: Failed in _eject(850,1.0) | {}'\
                    .format(msg)
                self.log.error(msg)
                return False, msg
            return True, 'eject(): Successfully finish!'

    def check_limitswitch(self, session, params=None):
        """check_limitswitch(io_name)

        **Task** - Print limit-switch ON/OFF state.

        Parameters:
            io_name (string): The IO name to be printed. Names configured in
                ``limitswitch_config.py``. If None, all limit-switches are
                printed.
        """
        if params is None:
            params = {}
        io_name = params.get('io_name', None)
        onoffs = []
        msg = ''
        with self.lock.acquire_timeout(timeout=3, job='check_limitswitch') \
                as acquired:
            if not acquired:
                self.log.warn(
                    'check_limitswitch(): '
                    'Lock could not be acquired because it is held by {}.'
                    .format(self.lock.job))
                return False,\
                    'check_limitswitch(): '\
                    'Could not acquire lock'

            onoffs = self.actuator.ls.get_onoff(io_name)
            io_names = self.actuator.ls.get_io_name(io_name)
            io_labels = self.actuator.ls.get_label(io_name)
            for i, io_name in enumerate(io_names):
                io_label = io_labels[i]
                msg += 'check_limitswitch(): {:10s} ({:20s}) : {}\n'\
                    .format(io_name, io_label, 'ON' if onoffs[i] else 'OFF')
            self.log.info(msg)
            return True, msg

    def check_stopper(self, session, params=None):
        """check_stopper(io_name)

        **Task** - Print stopper ON/OFF (ON: lock the actuator, OFF: release
        the actuator)

        Parameters:
            io_name (string): The IO name to be printed. Names configured in
                ``stopper_config.py``. If None, all stoppers are printed.
        """
        if params is None:
            params = {}
        io_name = params.get('io_name', None)
        onoffs = []
        msg = ''
        with self.lock.acquire_timeout(timeout=3, job='check_stopper') \
                as acquired:
            if not acquired:
                self.log.warn(
                    'check_stopper(): '
                    'Lock could not be acquired because it is held by {}.'
                    .format(self.lock.job))
                return False, 'check_stopper(): Could not acquire lock'

            onoffs = self.actuator.st.get_onoff(io_name)
            io_names = self.actuator.st.get_io_name(io_name)
            io_labels = self.actuator.st.get_label(io_name)
            for i, io_name in enumerate(io_names):
                io_label = io_labels[i]
                msg += 'check_stopper(): {:10s} ({:20s}) : {}\n'\
                    .format(io_name, io_label, 'ON' if onoffs[i] else 'OFF')
            self.log.info(msg)
            return True, msg

    def insert_homing(self, session, params=None):
        """insert_homing()

        **Task** - Insert slowly the wire-grid into the forebaffle interface
        above the SAT until the inside limit-switch becomes ON.

        """
        with self.lock.acquire_timeout(timeout=3, job='insert_homing')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'insert_homing(): '
                    'Lock could not be acquired because it is held by {}.'
                    .format(self.lock.job))
                return False, 'insert_homing(): Could not acquire lock'
            # Wait for a second before moving
            time.sleep(1)
            # Moving commands
            ret, msg = self._insert(1000, 0.1)
            if not ret:
                msg = 'insert_homing(): ERROR!: Failed '\
                      'in _insert(1000,0.1) | {}'.format(msg)
                self.log.error(msg)
                return False, msg
            return True, 'insert_homing(): Successfully finish!'

    def eject_homing(self, session, params=None):
        """eject_homing()

        **Task** - Eject slowly the wire-grid from the forebaffle interface
        above the SAT until the outside limit-switch becomes ON.

        """
        with self.lock.acquire_timeout(timeout=3, job='eject_homing')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'eject_homing(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'eject_homing(): Could not acquire lock'
            # Wait for a second before moving
            time.sleep(1)
            # Moving commands
            ret, msg = self._eject(1000, 0.1)
            if not ret:
                msg = 'eject_homing(): ERROR!: Failed '\
                      'in _eject(1000,0.1) | {}'.format(msg)
                self.log.error(msg)
                return False, msg
            return True, 'eject_homing(): Successfully finish!'

    @ocs_agent.param('distance', default=10., type=float)
    @ocs_agent.param('speedrate', default=0.2, type=float,
                     check=lambda x: 0.0 < x <= 1.0)
    def insert_test(self, session, params):
        """insert_test(distance=10, speedrate=0.1)

        **Task** - Insert slowly the wire-grid into the forebaffle interface
        above the SAT with a small distance.

        Parameters:
            distance (float): Actuator moving distance [mm] (default: 10)
            speedrate (float): Actuator speed rate [0.0, 1.0] (default: 0.2)
        """
        # Get parameters
        distance = params.get('distance')
        speedrate = params.get('speedrate')
        self.log.info('insert_test(): set distance   = {} mm'
                      .format(distance))
        self.log.info('insert_test(): set speed rate = {}'
                      .format(speedrate))

        with self.lock.acquire_timeout(timeout=3, job='insert_test') \
                as acquired:
            if not acquired:
                self.log.warn(
                    'insert_test(): '
                    'Lock could not be acquired because it is held by {}.'
                    .format(self.lock.job))
                return False, 'insert_test(): Could not acquire lock.'
            # Wait for a second before moving
            time.sleep(1)
            # Release the stoppers
            try:
                self.actuator.st.set_allon()
            except Exception as e:
                msg = 'insert_test(): ERROR!: '\
                      'Failed to run the stopper set_allon() '\
                      '--> Stop inserting! | Exception = "{}"'.format(e)
                self.log.error(msg)
                return False, msg
            # Moving commands
            ret, msg, LSonoff = self._forward(distance, speedrate=speedrate)
            if not ret:
                msg = 'insert_test(): ERROR!: Failed in _forward(10,1.) | {}'\
                    .format(msg)
                self.log.error(msg)
                return False, msg
            # Lock the stoppers
            self.actuator.st.set_alloff()
            return True, 'insert_test(): Successfully finish!'

    @ocs_agent.param('distance', default=10., type=float)
    @ocs_agent.param('speedrate', default=0.2, type=float,
                     check=lambda x: 0.0 < x <= 1.0)
    def eject_test(self, session, params):
        """eject_test(distance=10, speedrate=0.1)

        **Task** - Eject slowly the wire-grid from the forebaffle interface
        above the SAT with a small distance.

        Parameters:
            distance:  Actuator moving distance [mm] (default: 10)
            speedrate: Actuator speed rate [0.0, 1.0] (default: 0.2)
        """
        # Get parameters
        distance = params.get('distance', 10)
        speedrate = params.get('speedrate', 0.2)
        self.log.info('eject_test(): set distance   = {} mm'
                      .format(distance))
        self.log.info('eject_test(): set speed rate = {}'
                      .format(speedrate))

        with self.lock.acquire_timeout(timeout=3, job='eject_test')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'eject_test(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'eject_test(): Could not acquire lock.'
            # Wait for a second before moving
            time.sleep(1)
            # Release the stoppers
            try:
                self.actuator.st.set_allon()
            except Exception as e:
                msg = 'eject_test(): ERROR!: '\
                      'Failed to run the stopper set_allon() '\
                      '--> Stop ejecting! | Exception = "{}"'.format(e)
                self.log.error(msg)
                return False, msg
            # Moving commands
            ret, msg, LSonoff = self._backward(distance, speedrate=speedrate)
            if not ret:
                msg = 'eject_test(): ERROR!: Failed '\
                      'in _backward(10,1.) | {}'.format(msg)
                self.log.error(msg)
                return False, msg
            # Lock the stoppers
            self.actuator.st.set_alloff()
            return True, 'eject_test(): Successfully finish!'

    def motor_on(self, session, params=None):
        """motor_on()

        **Task** - Power ON the motors of the actuators.

        """
        with self.lock.acquire_timeout(timeout=3, job='motor_on')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'motor_on(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'motor_on(): Could not acquire lock.'
            self.log.warn('motor_on(): Try to power ON the actuator motors.')
            try:
                self.actuator.set_motor_onoff(onoff=True)
            except Exception as e:
                msg = 'motor_on(): ERROR!: '\
                      'Failed to power ON the actuator motors | '\
                      'Exception = "{}"'.format(e)
                self.log.error(msg)
                return False, msg
            return True, 'motor_on(): Successfully finish!'

    def motor_off(self, session, params=None):
        """motor_off()

        **Task** - Power OFF the motors of the actuators.

        """
        with self.lock.acquire_timeout(timeout=3, job='motor_off')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'motor_off(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'motor_off(): Could not acquire lock.'
            self.log.warn('motor_off(): Try to power OFF the actuator motors.')
            try:
                self.actuator.set_motor_onoff(onoff=False)
            except Exception as e:
                msg = 'motor_off(): ERROR!: '\
                      'Failed to power OFF the actuator motors | '\
                      'Exception = "{}"'.format(e)
                self.log.error(msg)
                return False, msg
            return True, 'motor_off(): Successfully finish!'

    def stop(self, session, params=None):
        """stop()

        **Task** - Emergency stop of the wire-grid actuator. (Disable the
        actuator motion.)

        .. note::
            This command can be excuted even if the other command is running.

        """
        self.log.warn('stop(): Try to stop and hold the actuator.')
        # This will disable move() command in Actuator class
        # until self.actuator.release() is called.
        ret = self.actuator.hold()
        if not ret:
            msg = 'stop(): ERROR!: Failed to hold the actuator'
            self.log.error(msg)
            return False, msg
        return True, 'stop(): Successfully finish!'

    def release(self, session, params=None):
        """release()

        **Task** - Enable the actuator moving.

        """
        with self.lock.acquire_timeout(timeout=3, job='release')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'release(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'release(): Could not acquire lock.'
            self.log.warn('release(): Try to release the actuator.')
            # This will enable move() command in Actuator class.
            try:
                self.actuator.release()
            except Exception as e:
                msg = 'release(): ERROR!: '\
                      'Failed to release the actuator | '\
                      'Exception = "{}"'.format(e)
                self.log.error(msg)
                return False, msg
            return True, 'release(): Successfully finish!'

    def reconnect(self, session, params=None):
        """reconnect()

        **Task** - Reconnect to the actuator controller.

        .. warning::
            This command turns OFF the motor power!

        """
        with self.lock.acquire_timeout(timeout=3, job='reconnect')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'reconnect(): '
                    'Lock could not be acquired because it is held by {}.'
                        .format(self.lock.job))
                return False, 'reconnect(): Could not acquire lock.'
            self.log.warn('reconnect(): *** Trying to reconnect... ***')
            # reconnect
            ret = self.actuator.reconnect()
            if not ret:
                msg = 'reconnect(): ERROR!: '\
                      'Failed to reconnect the actuator controller!'
                self.log.warn(msg)
                return False, msg
            # Check connection
            ret = self.actuator.check_connect()
            if ret:
                msg = 'reconnect(): Successfully reconnected to the actuator!'
                self.log.info(msg)
                return True, msg
            else:
                msg = 'reconnect(): ERROR!: '\
                      'Failed to reconnect to the actuator!'
                self.log.error(msg)
                return False, msg

    def acq(self, session, params=None):
        """acq()

        **Process** - Run data acquisition.

        Parameters:
           interval-time: interval time for data acquisition

        Notes:
            The most recent data collected is stored in session.data in the
            structure::

                >>> response.session['data']
                {'fields':
                    {
                     'motor':
                        0 or 1
                     'limitswitch':
                     {   'LSR1': 0 or 1, (0: OFF, 1:ON)
                         'LSR2': 0 or 1, (0: OFF, 1:ON)
                         .
                         .
                         },
                     'stopper':
                     {   'STR1': 0 or 1, (0: OFF, 1:ON)
                         'STR2': 0 or 1, (0: OFF, 1:ON)
                         .
                         .
                         }
                    },
                 'timestamp':1601925677.6914878
                }
        """
        if params is None:
            params = {}
        # Define data taking interval_time
        interval_time = params.get('interval-time', None)
        # If interval-time is None, use value passed to Agent init
        if interval_time is None:
            self.log.info(
                'acq(): '
                'Not set by parameter of "interval-time" for acq()')
            interval_time = self.interval_time
        else:
            try:
                interval_time = float(interval_time)
            except ValueError as e:
                self.log.warn(
                    'acq(): '
                    'Parameter of "interval-time" is incorrect | '
                    'Exception = "{}"'.format(e))
                interval_time = self.interval_time
        self.log.info(
            'acq(): '
            'interval time for acquisition of limit-switch & stopper = {} sec'
            .format(interval_time))

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            self.log.info('acq(): Start to take data')
            if not acquired:
                self.log.warn(
                    'acq(): '
                    'Lock could not be acquired because it is held by {}.'
                    .format(self.lock.job))
                return False, 'acq(): Could not acquire lock.'
            self.log.info('acq(): Got the lock')

            session.set_status('running')

            self.run_acq = True
            last_release = time.time()
            session.data = {'fields': {}}
            while self.run_acq:
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=180):
                        self.log.warn(
                            'acq(): '
                            'Could not re-acquire lock now held by {}.'
                            .format(self.lock.job))
                        return False,\
                            'acq(): Could not re-acquire lock (timeout)'

                current_time = time.time()
                data = {'timestamp': current_time,
                        'block_name': 'actuator_onoff',
                        'data': {}}

                # Take data
                onoff_mt = None
                onoff_dict_ls = {}
                onoff_dict_st = {}
                # Get onoff
                try:
                    onoff_mt = self.actuator.get_motor_onoff()
                    onoff_ls = self.actuator.ls.get_onoff()
                    onoff_st = self.actuator.st.get_onoff()
                except Exception as e:
                    msg = 'acq(): '\
                          'ERROR!: Failed to get status '\
                          'from the actuator controller!'\
                          '--> Stop acq()! | Exception = "{}"'.format(e)
                    self.log.error(msg)
                    return False, msg

                # Data for motor
                data['data']['motor'] = onoff_mt
                # Data for limitswitch
                for onoff, name in \
                        zip(onoff_ls, self.actuator.ls.io_names):
                    data['data']['limitswitch_{}'.format(name)] = onoff
                    onoff_dict_ls[name] = onoff
                # Data for stopper
                for onoff, name in \
                        zip(onoff_st, self.actuator.st.io_names):
                    data['data']['stopper_{}'.format(name)] = onoff
                    onoff_dict_st[name] = onoff
                # publish data
                self.agent.publish_to_feed('wgactuator', data)
                # store session.data
                field_dict = {'motor': onoff_mt,
                              'limitswitch': onoff_dict_ls,
                              'stopper': onoff_dict_st}
                session.data['timestamp'] = current_time
                session.data['fields'] = field_dict

                # wait an interval
                time.sleep(interval_time)
        # End of lock acquire

        self.agent.feeds['wgactuator'].flush_buffer()
        return True, 'acq(): Acquisition exited cleanly'

    def stop_acq(self, session, params=None):
        if self.run_acq:
            self.run_acq = False
            return True, 'stop_acq(): Stop data acquisition'
        return False, 'stop_acq(): acq is not currently running'

    # End of class WiregridActuatorAgent


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--interval-time', dest='interval_time',
                        type=float, default=1.,
                        help='Interval time for data taking')
    pgroup.add_argument('--ip-address', dest='ip_address',
                        type=str, default='192.168.1.100',
                        help='IP address of the actuator controller')
    pgroup.add_argument('--sleep', dest='sleep',
                        type=float, default=0.05,
                        help='Sleep time for every actuator command')
    pgroup.add_argument('--verbose', dest='verbose',
                        type=int, default=0,
                        help='Verbosity level')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='WiregridActuatorAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    actuator_agent = WiregridActuatorAgent(
        agent, ip_address=args.ip_address, interval_time=args.interval_time,
        sleep=args.sleep, verbose=args.verbose)

    agent.register_task('check_limitswitch', actuator_agent.check_limitswitch)
    agent.register_task('check_stopper', actuator_agent.check_stopper)
    agent.register_task('insert', actuator_agent.insert)
    agent.register_task('eject', actuator_agent.eject)
    agent.register_task('insert_homing', actuator_agent.insert_homing)
    agent.register_task('eject_homing', actuator_agent.eject_homing)
    agent.register_task('insert_test', actuator_agent.insert_test)
    agent.register_task('eject_test', actuator_agent.eject_test)
    agent.register_task('motor_on', actuator_agent.motor_on)
    agent.register_task('motor_off', actuator_agent.motor_off)
    agent.register_task('stop', actuator_agent.stop)
    agent.register_task('release', actuator_agent.release)
    agent.register_task('reconnect', actuator_agent.reconnect)
    agent.register_process('acq', actuator_agent.acq,
                           actuator_agent.stop_acq, startup=True)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import shutil
import socket
import subprocess
import threading
import time

import pytest
import serial


def create_device_emulator(responses, relay_type, port=9001, encoding='utf-8'):
    """Create a device emulator fixture.

    This provides a device emulator that can be used to mock a device during
    testing.

    Args:
        responses (dict): Dictionary with commands as keys, and responses as
            values. See :class:`.DeviceEmulator` for details.
        relay_type (str): Communication relay type. Either 'serial' or 'tcp'.
        port (int): Port for the TCP relay to listen for connections on.
            Defaults to 9001. Only used if relay_type is 'tcp'.
        encoding (str): Encoding for the messages and responses. See
            :func:`socs.testing.device_emulator.DeviceEmulator` for more
            details.

    Returns:
        function:
            A pytest fixture that creates a Device emulator of the specified
            type.

    """
    if relay_type not in ['serial', 'tcp']:
        raise NotImplementedError(f"relay_type '{relay_type}' is not"
                                  + "implemented or is an invalid type")

    @pytest.fixture()
    def create_device():
        device = DeviceEmulator(responses, encoding)

        if relay_type == 'serial':
            device.create_serial_relay()
        elif relay_type == 'tcp':
            device.create_tcp_relay(port)

        yield device

        device.shutdown()

    return create_device


class DeviceEmulator:
    """A mocked device which knows how to respond on various communication
    channels.

    Args:
        responses (dict): Initial responses, any response required by Agent
            startup, if any.
        encoding (str): Encoding for the messages and responses.
            DeviceEmulator will try to encode and decode messages with the
            given encoding. No encoding is used if set to None. That can be
            useful if you need to use raw data from your hardware. Defaults
            to 'utf-8'.

    Attributes:
        responses (dict): Current set of responses the DeviceEmulator would
            give. Should all be strings, not bytes-like.
        default_response (str): Default response to send if a command is
            unrecognized. No response is sent and an error message is logged if
            a command is unrecognized and the default response is set to None.
            Defaults to None.
        encoding (str): Encoding for the messages and responses, set by the
            encoding argument.
        _type (str): Relay type, either 'serial' or 'tcp'.
        _read (bool): Used to stop the background reading of data recieved on
            the relay.
        _conn (socket.socket): TCP connection for use in 'tcp' relay.

    """

    def __init__(self, responses, encoding='utf-8'):
        self.responses = responses
        self.default_response = None
        self.encoding = encoding
        self._type = None
        self._read = True
        self._conn = None

    @staticmethod
    def _setup_socat():
        """Setup a data relay with socat.

        The "./responder" link is the external end of the relay, which the Agent
        should connect to. "./internal" is used within the DeviceEmulator object to accept
        commands and to respond to the Agent.

        """
        socat = shutil.which('socat')
        cmd = [socat, '-d', '-d', 'pty,link=./responder,b57600', 'pty,link=./internal,b57600']
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        time.sleep(1)

        return proc

    def create_serial_relay(self):
        """Create the serial relay, emulating a hardware device connected over
        serial.

        This first uses ``socat`` to setup a relay. It then connects to the
        "internal" end of the relay, ready to receive communications sent to the
        "responder" end of the relay. This end of the relay is located at
        ``./responder``. You will need to configure your Agent to use that path for
        communication.

        Next it creates a thread to read commands sent to the serial relay in
        the background. This allows responses to be defined within a test using
        DeviceEmulator.define_responses() after instantiation of the DeviceEmulator
        object within a given test.

        """
        self._type = 'serial'
        self.proc = self._setup_socat()
        self.ser = serial.Serial(
            './internal',
            baudrate=57600,
            timeout=5,
        )
        bkg_read = threading.Thread(name='background',
                                    target=self._read_serial)
        bkg_read.start()

    def _get_response(self, msg):
        """Determine the response to a given message.

        Args:
            msg (str): Command string to get the response for.

        Returns:
            str: Response string. Will return None if a valid response is not
                 found.

        """
        if self.responses is None:
            return

        if msg not in self.responses and self.default_response is not None:
            return self.default_response

        try:
            if isinstance(self.responses[msg], list):
                response = self.responses[msg].pop(0)
            else:
                response = self.responses[msg]
        except Exception as e:
            print(f"encountered error {e}")
            response = None

        return response

    def _read_serial(self):
        """Loop until shutdown, reading any commands sent over the relay.
        Respond immediately to a command with the response in self.responses.

        """
        self._read = True

        while self._read:
            if self.ser.in_waiting > 0:
                msg = self.ser.readline()
                if self.encoding:
                    msg = msg.strip().decode(self.encoding)
                print(f"msg='{msg}'")

                response = self._get_response(msg)

                # Avoid user providing bytes-like response
                if isinstance(response, bytes) and self.encoding is not None:
                    response = response.decode()

                if response is None:
                    continue

                print(f"response='{response}'")
                if self.encoding:
                    response = (response + '\r\n').encode(self.encoding)
                self.ser.write(response)

            time.sleep(0.01)

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        """Shutdown communication on the configured relay. This will stop any
        attempt to read communication on the relay, as well as shutdown the relay
        itself.

        """
        # print('shutting down background reading')
        self._read = False
        time.sleep(1)
        if self._type == 'serial':
            # print('shutting down socat relay')
            self.proc.terminate()
            out, err = self.proc.communicate()
            # print(out, err)
        if self._type == 'tcp':
            # print('shutting down background tcp relay')
            if self._conn:
                self._conn.close()
                self._sock.close()

    def _read_socket(self, port):
        """Loop until shutdown, reading any commands sent over the relay.
        Respond immediately to a command with the response in self.responses.

        Args:
            port (int): Port for the TCP relay to listen for connections on.

        """
        self._read = True

        # Listen for connections
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.bind(('127.0.0.1', port))
        self._sock.listen(1)
        print("Device emulator waiting for tcp client connection")
        self._conn, client_address = self._sock.accept()
        print(f"Client connection made from {client_address}")

        while self._read:
            try:
                msg = self._conn.recv(4096)
            # Was seeing this on tests in the cryomech agent
            except ConnectionResetError:
                print('Caught connection reset on Agent clean up')
                break
            if self.encoding:
                msg = msg.strip().decode(self.encoding)
            if msg:
                print(f"msg='{msg}'")

                response = self._get_response(msg)

                # Avoid user providing bytes-like response
                if isinstance(response, bytes) and self.encoding is not None:
                    response = response.decode()

                if response is None:
                    continue

                print(f"response='{response}'")
                if self.encoding:
                    response = response.encode(self.encoding)
                self._conn.sendall(response)

            time.sleep(0.01)

        self._conn.close()
        self._sock.close()

    def create_tcp_relay(self, port):
        """Create the TCP relay, emulating a hardware device connected over
        TCP.

        Creates a thread to read commands sent to the TCP relay in the
        background. This allows responses to be defined within a test using
        DeviceEmulator.define_responses() after instantiation of the
        DeviceEmulator object within a given test.

        Args:
            port (int): Port for the TCP relay to listen for connections on.

        """
        self._type = 'tcp'
        bkg_read = threading.Thread(name='background',
                                    target=self._read_socket,
                                    kwargs={'port': port})
        bkg_read.start()

    def define_responses(self, responses, default_response=None):
        """Define what responses are available to reply with on the configured
        communication relay.

        Args:
            responses (dict): Dictionary of commands: response. Values can be a
                list, in which case the responses in the list are popped and
                given in order until depleted.
            default_response (str): Default response to send if a command is
                unrecognized. No response is sent and an error message is
                logged if a command is unrecognized and the default response is
                set to None. Defaults to None.

        Examples:
            The given responses might look like::

                >>> responses = {'KRDG? 1': '+1.7E+03'}
                >>> responses = {'*IDN?': 'LSCI,MODEL425,4250022,1.0',
                                 'RDGFIELD?': ['+1.0E-01', '+1.2E-01', '+1.4E-01']}

        Notes:
            The DeviceEmulator will handle encoding/decoding. The responses
            defined should all be strings, not bytes-like, unless you set
            ``encoding=None``.

        """
        print(f"responses set to {responses}")
        self.responses = responses
        print(f"default response set to '{default_response}'")
        self.default_response = default_response



"""
Register our agents in ocs central.  In order for this script to
be imported by site_config.scan_for_agents(), it must be in the python
path and called something like ocs_plugin_*.
"""

import os

import ocs

root = os.path.abspath(os.path.split(__file__)[0])

for n, f in [
        ('ACUAgent', 'acu/agent.py'),
        ('BlueforsAgent', 'bluefors/agent.py'),
        ('CrateAgent', 'smurf_crate_monitor/agent.py'),
        ('CryomechCPAAgent', 'cryomech_cpa/agent.py'),
        ('FPGAAgent', 'holo_fpga/agent.py'),
        ('FTSAerotechAgent', 'fts_aerotech/agent.py'),
        ('HWPBBBAgent', 'hwp_encoder/agent.py'),
        ('HWPPicoscopeAgent', 'hwp_picoscope/agent.py'),
        ('ibootbarAgent', 'ibootbar/agent.py'),
        ('LabJackAgent', 'labjack/agent.py'),
        ('Lakeshore240Agent', 'lakeshore240/agent.py'),
        ('Lakeshore336Agent', 'lakeshore336/agent.py'),
        ('Lakeshore370Agent', 'lakeshore370/agent.py'),
        ('Lakeshore372Agent', 'lakeshore372/agent.py'),
        ('Lakeshore425Agent', 'lakeshore425/agent.py'),
        ('LATRtXYStageAgent', 'xy_stage/agent.py'),
        ('MagpieAgent', 'magpie/agent.py'),
        ('MeinbergM1000Agent', 'meinberg_m1000/agent.py'),
        ('PfeifferAgent', 'pfeiffer_tpg366/agent.py'),
        ('PfeifferTC400Agent', 'pfeiffer_tc400/agent.py'),
        ('PysmurfController', 'pysmurf_controller/agent.py'),
        ('PysmurfMonitor', 'pysmurf_monitor/agent.py'),
        ('RotationAgent', 'hwp_rotation/agent.py'),
        ('ScpiPsuAgent', 'scpi_psu/agent.py'),
        ('SmurfFileEmulator', 'smurf_file_emulator/agent.py'),
        ('SmurfStreamSimulator', 'smurf_stream_simulator/agent.py'),
        ('SupRsync', 'suprsync/agent.py'),
        ('SynaccessAgent', 'synacc/agent.py'),
        ('SynthAgent', 'holo_synth/agent.py'),
        ('TektronixAWGAgent', 'tektronix3021c/agent.py'),
        ('ThorlabsMC2000BAgent', 'thorlabs_mc2000b/agent.py'),
        ('VantagePro2Agent', 'vantagepro2/agent.py'),
        ('WiregridActuatorAgent', 'wiregrid_actuator/agent.py'),
        ('WiregridEncoderAgent', 'wiregrid_encoder/agent.py'),
        ('WiregridKikusuiAgent', 'wiregrid_kikusui/agent.py'),
]:
    ocs.site_config.register_agent_class(n, os.path.join(root, f))



from setuptools import find_packages, setup

import versioneer

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Optional Dependencies
# ACU Agent
# acu_deps = ['soaculib @ git+https://github.com/simonsobs/soaculib.git@master']

# Holography FPGA and Synthesizer Agents
# holography_deps = [  # Note: supports python 3.8 only!
#     'casperfpga @ git+https://github.com/casper-astro/casperfpga.git@py38',
#     'holog_daq @ git+https://github.com/McMahonCosmologyGroup/holog_daq.git@main',
# ]

# Labjack Agent
labjack_deps = [
    'labjack-ljm',
    'numexpr',
    'scipy',
]

# Magpie Agent
magpie_deps = [
    'pandas',
    'scipy',
    'so3g',
]

# Pfeiffer TC 400 Agent
pfeiffer_deps = ['pfeiffer-vacuum-protocol==0.4']

# Pysmurf Controller Agent
# pysmurf_deps = [
#     'pysmurf @ git+https://github.com/slaclab/pysmurf.git@main',
#     'sodetlib @ git+https://github.com/simonsobs/sodetlib.git@master',
#     'sotodlib @ git+https://github.com/simonsobs/sotodlib.git@master',
# ]

# SMuRF File Emulator, SMuRF Stream Simulator
smurf_sim_deps = ['so3g']

# Synaccess Agent
synacc_deps = ['requests']

# LATRt XY Stage Agent
# xy_stage_deps = [
#     'xy_stage_control @ git+https://github.com/kmharrington/xy_stage_control.git@main',
# ]

# Note: Not including the holograph deps, which are Python 3.8 only
# all_deps = acu_deps + labjack_deps + magpie_deps + pfeiffer_deps + \
#    pysmurf_deps + smurf_sim_deps + synacc_deps + xy_stage_deps
all_deps = labjack_deps + magpie_deps + pfeiffer_deps + \
    smurf_sim_deps + synacc_deps
all_deps = list(set(all_deps))

setup(
    name='socs',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Simons Observatory Control System',
    package_dir={'socs': 'socs'},
    packages=find_packages(),
    package_data={'socs': [
        'agents/smurf_file_emulator/*.yaml',
        'agents/labjack/cal_curves/*.txt',
    ]},
    entry_points={
        'ocs.plugins': [
            'socs = socs.plugin',
        ],
    },
    url="https://github.com/simonsobs/socs",
    project_urls={
        "Source Code": "https://github.com/simonsobs/ocs",
        "Documentation": "https://ocs.readthedocs.io/",
        "Bug Tracker": "https://github.com/simonsobs/ocs/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Framework :: Twisted",
    ],
    python_requires=">=3.7",
    install_requires=[
        'autobahn[serialization]',
        'numpy',
        'ocs',
        'pyserial',
        'pysnmp',
        'pysmi',
        'pyyaml',
        'sqlalchemy>=1.4',
        'twisted',
    ],
    extras_require={
        'all': all_deps,
        # 'acu': acu_deps,
        # 'holography': holography_deps,
        'labjack': labjack_deps,
        'magpie': magpie_deps,
        'pfeiffer': pfeiffer_deps,
        # 'pysmurf': pysmurf_deps,
        'smurf_sim': smurf_sim_deps,
        'synacc': synacc_deps,
        # 'xy_stage': xy_stage_deps,
    },
)



import sys
import time

import serial

from socs.common import moxa_serial as mx


class PMX:
    """The PMX object is for communicating with the Kikusui PMX power supplies.

    Args:
        rtu_port (str): Serial RTU port
        tcp_ip (str): TCP IP address
        tcp_port (int): TCP port
        timeout (int): Connection timeout
    """

    def __init__(self, rtu_port=None, tcp_ip=None, tcp_port=None, timeout=None):
        # Connect to device
        self.using_tcp = None
        self._rtu_port = None
        self.ser = None
        msg = self.__conn(rtu_port, tcp_ip, tcp_port, timeout)
        print(msg)
        self._remote_Mode()

        # Timing variables
        self._tstep = 0.1  # sec

    def __del__(self):
        if not self.using_tcp:
            print(
                "Disconnecting from RTU port %s"
                % (self._rtu_port))
            if self.ser:
                self.ser.close()
        else:
            print(
                "Disconnecting from TCP IP %s at port %d"
                % (self._tcp_ip, self._tcp_port))
        return

    def check_connect(self):
        """Check the connection."""
        try:
            if not self.using_tcp:
                self.ser.inWaiting()
            else:
                self.clean_serial()
                self.wait()
                self.ser.write("OUTP?\n\r")
                self.wait()
                val = (self.ser.readline().strip())
                val = int(val)
        except Exception as e:
            msg = 'Could not connect to the PMX serial! | Error: "{}"'.format(e)
            return msg, False
        return 'Successfully connect to the PMX serial.', True

    def check_voltage(self):
        """Check the voltage."""
        self.clean_serial()
        self.ser.write("MEAS:VOLT?\n\r")
        self.wait()
        try:
            val = float(self.ser.readline())
            msg = "Measured voltage = %.3f V" % (val)
            # print(msg)
        except ValueError:
            val = -999.
            msg = 'WARNING! Could not get correct voltage value! | Response = "%s"' % (val)
            print(msg)
        return msg, val

    def check_current(self):
        """Check the current."""
        self.clean_serial()
        self.ser.write("MEAS:CURR?\n\r")
        self.wait()
        try:
            val = float(self.ser.readline())
            msg = "Measured current = %.3f A" % (val)
            # print(msg)
        except ValueError:
            val = -999.
            msg = 'WARNING! Could not get correct current value! | Response = "%s"' % (val)
            print(msg)
        return msg, val

    def check_voltage_current(self):
        """Check both the voltage and current."""
        self.clean_serial()
        voltage = self.check_voltage()[1]
        current = self.check_current()[1]
        # msg = (
        #     "Measured voltage = %.3f V\n"
        #     "Measured current = %.3f A\n"
        #     % (voltage, current))
        # print(msg)
        return voltage, current

    def check_voltage_setting(self):
        """ Check the voltage setting """
        self.clean_serial()
        for i in range(10):
            self.ser.write("VOLT?\n\r")
            self.wait()
            val = (self.ser.readline().strip())
            if len(val) > 0:
                break
        try:
            val = float(val)
            msg = "Voltage setting = %.3f V" % (val)
            # print(msg)
        except ValueError:
            val = -999.
            msg = 'WARNING! Could not get correct voltage-setting value! | Response = "%s"' % (val)
            print(msg)
        return msg, val

    def check_current_setting(self):
        """ Check the current setting """
        self.clean_serial()
        for i in range(10):
            self.ser.write("CURR?\n\r")
            self.wait()
            val = (self.ser.readline().strip())
            if len(val) > 0:
                break
        try:
            val = float(val)
            msg = "Current setting = %.3f A" % (val)
            # print(msg)
        except ValueError:
            val = -999.
            msg = 'WARNING! Could not get correct current-setting value! | Response = "%s"' % (val)
            print(msg)
        return msg, val

    def check_voltage_current_setting(self):
        """ Check both the voltage and current setting """
        self.clean_serial()
        voltage = self.check_voltage_setting()[1]
        current = self.check_current_setting()[1]
        msg = (
            "Voltage setting = %.3f V\n"
            "Current setting = %.3f A\n"
            % (voltage, current))
        # print(msg)
        return msg, voltage, current

    def check_output(self):
        """Return the output status."""
        self.clean_serial()
        self.ser.write("OUTP?\n\r")
        self.wait()
        try:
            val = int(self.ser.readline())
        except ValueError:
            val = -999
            msg = 'WARNING! Could not get correct output value! | Response = "%s"' % (val)
            print(msg)
            return msg, val
        if val == 0:
            msg = "Measured output state = OFF"
        elif val == 1:
            msg = "Measured output state = ON"
        else:
            msg = "Failed to measure output..."
        # print(msg)
        return msg, val

    def set_voltage(self, val, silent=False):
        """Set the PMX voltage."""
        self.clean_serial()
        self.ser.write("VOLT %f\n\r" % (float(val)))
        self.wait()
        self.ser.write("VOLT?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "Voltage set = %.3f V" % (float(val))
        if silent is not True:
            print(msg)

        return msg

    def set_current(self, val, silent=False):
        """Set the PMX on."""
        self.clean_serial()
        self.ser.write("CURR %f\n\r" % (float(val)))
        self.wait()
        self.ser.write("CURR?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "Current set = %.3f A\n" % (float(val))
        if silent is not True:
            print(msg)

        return msg

    def use_external_voltage(self):
        """Set PMX to use external voltage."""
        self.clean_serial()
        self.ser.write("VOLT:EXT:SOUR VOLT\n\r")
        self.wait()
        self.ser.write("VOLT:EXT:SOUR?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "External source = %s" % (val)
        print(msg)

        return msg

    def ign_external_voltage(self):
        """Set PMX to ignore external voltage."""
        self.clean_serial()
        self.ser.write("VOLT:EXT:SOUR NONE\n\r")
        self.wait()
        self.ser.write("VOLT:EXT:SOUR?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "External source = %s" % (val)
        print(msg)

        return msg

    def set_voltage_limit(self, val, silent=False):
        """Set the PMX voltage limit."""
        self.clean_serial()
        self.ser.write("VOLT:PROT %f\n\r" % (float(val)))
        self.wait()
        self.ser.write("VOLT:PROT?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "Voltage limit set = %.3f V" % (float(val))
        if silent is not True:
            print(msg)

        return msg

    def set_current_limit(self, val, silent=False):
        """Set the PMX current limit."""
        self.clean_serial()
        self.ser.write("CURR:PROT %f\n\r" % (float(val)))
        self.wait()
        self.ser.write("CURR:PROT?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "Current limit set = %.3f A\n" % (float(val))
        if silent is not True:
            print(msg)

        return msg

    def turn_on(self):
        """Turn the PMX on."""
        self.clean_serial()
        self.ser.write("OUTP ON\n\r")
        self.wait()
        self.ser.write("OUTP?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "Output state = %s" % (val)
        print(msg)

        return msg

    def turn_off(self):
        """Turn the PMX off."""
        self.clean_serial()
        self.ser.write("OUTP OFF\n\r")
        self.wait()
        self.ser.write("OUTP?\n\r")
        self.wait()
        val = self.ser.readline()
        msg = "Output state = %s" % (val)
        print(msg)

        return msg

    # ***** Helper Methods *****
    def __conn(self, rtu_port=None, tcp_ip=None, tcp_port=None, timeout=None):
        """Connect to the PMX module.

        Args:
            rtu_port (str): Serial RTU port
            tcp_ip (str): TCP IP address
            tcp_port (int): TCP port
            timeout (int): Connection timeout
        """
        if rtu_port is None and (tcp_ip is None or tcp_port is None):
            raise Exception(
                "Aborted PMX._conn() due to no RTU or "
                "TCP port specified")
        elif (rtu_port is not None
              and (tcp_ip is not None or tcp_port is not None)):
            raise Exception(
                "Aborted PMX._conn() due to RTU and TCP port both being "
                "specified. Can only have one or the other.")
        elif rtu_port is not None:
            self.ser = serial.Serial(
                port=rtu_port, baudrate=19200, bytesize=8,
                parity='N', stopbits=1, timeout=timeout)
            self._rtu_port = rtu_port
            self.using_tcp = False
            msg = "Connected to RTU port %s" % (rtu_port)
        elif tcp_ip is not None and tcp_port is not None:
            self.ser = mx.Serial_TCPServer((tcp_ip, tcp_port), timeout)
            self._tcp_ip = tcp_ip
            self._tcp_port = int(tcp_port)
            self.using_tcp = True
            msg = "Connected to TCP IP %s at port %d" % (tcp_ip, tcp_port)
        else:
            raise Exception(
                "Aborted PMX._conn() due to unknown error")
        return msg

    def wait(self):
        """Sleep."""
        time.sleep(0.05)
        return True

    def clean_serial(self):
        """Flush the serial buffer."""
        # if not False:
        #    self.ser.reset_input_buffer()
        #    self.ser.reset_output_buffer()
        #    self.ser.flush()
        # else:
        # self.ser.flushInput()
        self.ser.flushInput()
        return True

    def _remote_Mode(self):
        """Enable remote control."""
        self.clean_serial()
        self.ser.write('SYST:REM\n\r')
        self.wait()
        return True


class Command:
    """The Command object is used to command the PMX.

    Args:
        PMX (PMX): PMX object

    """

    def __init__(self, input_PMX):
        # PMX connection
        if input_PMX is not None:
            self._PMX = input_PMX
        else:
            raise Exception(
                "PMX object not passed to Command constructor\n")

        # Dict of commands
        self._cmds = {
            "set_port": "P",
            "check_v": "V?",
            "check_c": "C?",
            "check_vc": "VC?",
            "check_vs": "VS?",
            "check_cs": "CS?",
            "check_vcs": "VCS?",
            "check_out": "O?",
            "set_v": "V",
            "set_c": "C",
            "set_v_lim": "VL",
            "set_c_lim": "CL",
            "set_on": "ON",
            "set_off": "OFF",
            "get_help": "H",
            "use_ext": "U",
            "ign_ext": "I",
            "stop": "Q"}

    def get_help(self):
        """Print possible commands."""
        wrstr = (
            "\nChange ttyUSB port = '%s'\n"
            "Check output voltage = '%s'\n"
            "Check output current = '%s'\n"
            "Check output voltage and current = '%s'\n"
            "Check voltage setting = '%s'\n"
            "Check current setting = '%s'\n"
            "Check voltage and current setting = '%s'\n"
            "Check output state = '%s'\n"
            "Set output voltage = '%s' [setting]\n"
            "Set output current = '%s' [setting]\n"
            "Set output voltage limit = '%s'\n"
            "Set output current limit = '%s'\n"
            "Turn output on = '%s'\n"
            "Turn output off = '%s'\n"
            "Print possible commands = '%s'\n"
            "Use external voltage = '%s'\n"
            "Ignore external voltage = '%s'\n"
            "Quit program = '%s'\n"
            % (self._cmds["set_port"],
               self._cmds["check_v"],
               self._cmds["check_c"],
               self._cmds["check_vc"],
               self._cmds["check_vs"],
               self._cmds["check_cs"],
               self._cmds["check_vcs"],
               self._cmds["check_out"],
               self._cmds["set_v"],
               self._cmds["set_c"],
               self._cmds["set_v_lim"],
               self._cmds["set_c_lim"],
               self._cmds["set_on"],
               self._cmds["set_off"],
               self._cmds["get_help"],
               self._cmds["use_ext"],
               self._cmds["ign_ext"],
               self._cmds["stop"]))
        return wrstr

    def user_input(self, arg):
        """Take user input and execute PMX command."""
        argv = arg.split()
        # if len(args) > 0:
        # value = float(args[0])

        while len(argv):
            cmd = str(argv.pop(0)).upper()
            # No command
            if cmd == '':
                return
            # Check voltage
            elif cmd == self._cmds["check_v"]:
                return self._PMX.check_voltage()
            # Check current
            elif cmd == self._cmds["check_c"]:
                return self._PMX.check_current()
            # Check voltage and current
            elif cmd == self._cmds["check_vc"]:
                return self._PMX.check_voltage_current()
            # Check voltage setting
            elif cmd == self._cmds["check_vs"]:
                return self._PMX.check_voltage_setting()
            # Check current setting
            elif cmd == self._cmds["check_cs"]:
                return self._PMX.check_current_setting()
            # Check voltage and current
            elif cmd == self._cmds["check_vcs"]:
                return self._PMX.check_voltage_current_setting()
            # Check output state
            elif cmd == self._cmds["check_out"]:
                return self._PMX.check_output()
            # Turn output state ON
            elif cmd == self._cmds["set_on"]:
                return self._PMX.turn_on()
            # Turn output state OFF
            elif cmd == self._cmds["set_off"]:
                return self._PMX.turn_off()
            # Get HELP

            # elif cmd == self._cmds["set_v"]:
                # print(value)
                # ret = self._PMX.set_voltage(value)

            # elif cmd == self._cmds["set_c"]:
                # ret = self._PMX.set_current(value)

            elif cmd == self._cmds["use_ext"]:
                return self._PMX.use_external_voltage()
            elif cmd == self._cmds["ign_ext"]:
                return self._PMX.ign_external_voltage()
            elif cmd == self._cmds["get_help"]:
                ret = self.get_help()
                print(ret)
            # Exit the program
            elif cmd == self._cmds["stop"]:
                sys.exit("\nExiting...")
            # Set the RTU port
            elif cmd == self._cmds["set_port"]:
                if self._PMX.using_tcp:
                    print("Connected via TCP rather than RTU. Cannot set RTU port")
                    return False
                set_val = self._int(argv.pop(0))
                if set_val is not None:
                    del self._PMX
                    self._PMX = PMX(set_val)
                else:
                    return False
            elif cmd == self._cmds["set_v"]:
                set_val = self._float(argv.pop(0))
                if set_val is not None:
                    self._PMX.set_voltage(set_val)
                else:
                    return False
            elif cmd.lower() == self._cmds["set_c"].lower():
                set_val = self._float(argv.pop(0))
                if set_val is not None:
                    self._PMX.set_current(set_val)
                else:
                    return False
            elif cmd == self._cmds["set_v_lim"]:
                set_val = self._float(argv.pop(0))
                if set_val is not None:
                    self._PMX.set_voltage_limit(set_val)
                else:
                    return False
            elif cmd.lower() == self._cmds["set_c_lim"].lower():
                set_val = self._float(argv.pop(0))
                if set_val is not None:
                    self._PMX.set_current_limit(set_val)
                else:
                    return False
            else:
                print("Command '%s' not understood..." % (cmd))
                return False
        return True

    # ***** Helper Methods *****
    def _float(self, val):
        """Try to convert a value to a float."""
        try:
            return float(val)
        except ValueError:
            print("Input '%s' not understood, must be a float..." % (val))
            return None

    def _int(self, val):
        """Try to convert a value to an int."""
        try:
            return int(val)
        except ValueError:
            print("Input '%s' not understood, must be a int..." % (val))
            return None



import argparse
import os
import time
import warnings
from typing import Optional

import txaio

from socs.Lakeshore.Lakeshore240 import Module

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    from ocs import ocs_agent, site_config
    from ocs.ocs_twisted import TimeoutLock


class LS240_Agent:

    def __init__(self, agent, port="/dev/ttyUSB0", f_sample=2.5):

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.port = port
        self.module: Optional[Module] = None

        self.f_sample = f_sample

        self.initialized = False
        self.take_data = False

        # Registers Temperature and Voltage feeds
        agg_params = {
            'frame_length': 60,
        }
        self.agent.register_feed('temperatures',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    # Task functions.
    def init_lakeshore(self, session, params=None):
        """init_lakeshore(auto_acquire=False)

        **Task** - Perform first time setup of the Lakeshore 240 Module.

        Parameters:
            auto_acquire (bool, optional): Starts data acquisition after
                initialization if True. Defaults to False.

        """
        if params is None:
            params = {}

        auto_acquire = params.get('auto_acquire', False)

        if self.initialized:
            return True, "Already Initialized Module"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('starting')

            self.module = Module(port=self.port)
            print("Initialized Lakeshore module: {!s}".format(self.module))
            session.add_message("Lakeshore initialized with ID: %s" % self.module.inst_sn)

        self.initialized = True

        # Start data acquisition if requested
        if auto_acquire:
            self.agent.start('acq')

        return True, 'Lakeshore module initialized.'

    def set_values(self, session, params=None):
        """set_values(channel, sensor=None, auto_range=None, range=None,\
                current_reversal=None, units=None, enabled=None, name=None)

        **Task** - Set sensor parameters for a Lakeshore240 Channel.

        Args:
            channel (int):
                Channel number to set. Valid choices are 1-8.
            sensor (int, optional):
                Specifies sensor type.  See
                :func:`socs.Lakeshore.Lakeshore240.Channel.set_values` for
                possible types.
            auto_range (int, optional):
                Specifies if channel should use autorange. Must be 0 or 1.
            range (int, optional):
                Specifies range if auto_range is false. Only settable for NTC
                RTD.  See
                :func:`socs.Lakeshore.Lakeshore240.Channel.set_values` for
                possible ranges.
            current_reversal (int, optional):
                Specifies if input current reversal is on or off.
                Always 0 if input is a diode.
            units (int, optional):
                Specifies preferred units parameter, and sets the units for
                alarm settings.  See
                :func:`socs.Lakeshore.Lakeshore240.Channel.set_values` for
                possible units.
            enabled (int, optional):
                Sets if channel is enabled.
            name (str, optional):
                Sets name of channel.

        """
        if params is None:
            params = {}

        with self.lock.acquire_timeout(0, job='set_values') as acquired:
            if not acquired:
                self.log.warn("Could not start set_values because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            self.module.channels[params['channel'] - 1].set_values(
                sensor=params.get('sensor'),
                auto_range=params.get('auto_range'),
                range=params.get('range'),
                current_reversal=params.get('current_reversal'),
                unit=params.get('unit'),
                enabled=params.get('enabled'),
                name=params.get('name'),
            )

        return True, 'Set values for channel {}'.format(params['channel'])

    def upload_cal_curve(self, session, params=None):
        """upload_cal_curve(channel, filename)

        **Task** - Upload a calibration curve to a channel.

        Parameters:
            channel (int): Channel number, 1-8.
            filename (str): Filename for calibration curve.

        """
        channel = params['channel']
        filename = params['filename']

        with self.lock.acquire_timeout(0, job='upload_cal_curve') as acquired:
            if not acquired:
                self.log.warn("Could not start set_values because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            channel = self.module.channels[channel - 1]
            self.log.info("Starting upload to channel {}...".format(channel))
            channel.load_curve(filename)
            self.log.info("Finished uploading.")

        return True, "Uploaded curve to channel {}".format(channel)

    def acq(self, session, params=None):
        """acq(sampling_frequency=2.5)

        **Process** - Start data acquisition.

        Parameters:
            sampling_frequency (float):
                Sampling frequency for data collection. Defaults to 2.5 Hz


        The most recent data collected is stored in session data in the
        structure::

            >>> response.session['data']
            {"fields":
                {"Channel_1": {"T": 99.26, "V": 99.42},
                 "Channel_2": {"T": 99.54, "V": 101.06},
                 "Channel_3": {"T": 100.11, "V":100.79},
                 "Channel_4": {"T": 98.49, "V": 100.77},
                 "Channel_5": {"T": 97.75, "V": 101.45},
                 "Channel_6": {"T": 99.58, "V": 101.75},
                 "Channel_7": {"T": 98.03, "V": 100.82},
                 "Channel_8": {"T": 101.14, "V":101.01}},
             "timestamp":1601925677.6914878}

        """
        if params is None:
            params = {}

        f_sample = params.get('sampling_frequency')
        # If f_sample is None, use value passed to Agent init
        if f_sample is None:
            f_sample = self.f_sample

        sleep_time = 1 / f_sample - 0.01

        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already running"
                              .format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')

            self.take_data = True

            session.data = {"fields": {}}

            while self.take_data:
                current_time = time.time()
                data = {
                    'timestamp': current_time,
                    'block_name': 'temps',
                    'data': {}
                }

                for chan in self.module.channels:
                    # Read sensor on channel
                    chan_string = "Channel_{}".format(chan.channel_num)
                    temp_reading = chan.get_reading(unit='K')
                    sensor_reading = chan.get_reading(unit='S')

                    # For data feed
                    data['data'][chan_string + '_T'] = temp_reading
                    data['data'][chan_string + '_V'] = sensor_reading

                    # For session.data
                    field_dict = {chan_string: {"T": temp_reading, "V": sensor_reading}}
                    session.data['fields'].update(field_dict)

                self.agent.publish_to_feed('temperatures', data)

                session.data.update({'timestamp': current_time})

                time.sleep(sleep_time)

            self.agent.feeds['temperatures'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--serial-number', type=str,
                        help="Serial number of your Lakeshore240 device")
    pgroup.add_argument('--port', type=str,
                        help="Path to USB node for the lakeshore")
    pgroup.add_argument('--mode', type=str, choices=['idle', 'init', 'acq'],
                        help="Starting action for the agent.")
    pgroup.add_argument('--sampling-frequency', type=float,
                        help="Sampling frequency for data acquisition")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Not used anymore, but we don't it to break the agent if these args are passed
    parser.add_argument('--fake-data', help=argparse.SUPPRESS)
    parser.add_argument('--num-channels', help=argparse.SUPPRESS)

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class='Lakeshore240Agent',
                                  parser=parser,
                                  args=args)

    if args.fake_data is not None:
        warnings.warn("WARNING: the --fake-data parameter is deprecated, please "
                      "remove from your site-config file", DeprecationWarning)

    if args.num_channels is not None:
        warnings.warn("WARNING: the --num-channels parameter is deprecated, please "
                      "remove from your site-config file", DeprecationWarning)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    device_port = None
    if args.port is not None:
        device_port = args.port
    else:  # Tries to find correct USB port automatically

        # This exists if udev rules are setup properly for the 240s
        if os.path.exists('/dev/{}'.format(args.serial_number)):
            device_port = "/dev/{}".format(args.serial_number)

        elif os.path.exists('/dev/serial/by-id'):
            ports = os.listdir('/dev/serial/by-id')
            for port in ports:
                if args.serial_number in port:
                    device_port = "/dev/serial/by-id/{}".format(port)
                    print("Found port {}".format(device_port))
                    break

    if device_port is None:
        print("Could not find device port for {}".format(args.serial_number))
        return

    agent, runner = ocs_agent.init_site_agent(args)

    kwargs = {
        'port': device_port
    }

    if args.sampling_frequency is not None:
        kwargs['f_sample'] = float(args.sampling_frequency)

    therm = LS240_Agent(agent, **kwargs)

    agent.register_task('init_lakeshore', therm.init_lakeshore,
                        startup=init_params)
    agent.register_task('set_values', therm.set_values)
    agent.register_task('upload_cal_curve', therm.upload_cal_curve)
    agent.register_process('acq', therm.acq, therm._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



package_name = 'socs'
agents = {
    'ACUAgent': {'module': 'socs.agents.acu.agent', 'entry_point': 'main'},
    'BlueforsAgent': {'module': 'socs.agents.bluefors.agent', 'entry_point': 'main'},
    'CrateAgent': {'module': 'socs.agents.smurf_crate_monitor.agent', 'entry_point': 'main'},
    'CryomechCPAAgent': {'module': 'socs.agents.cryomech_cpa.agent', 'entry_point': 'main'},
    'FPGAAgent': {'module': 'socs.agents.holo_fpga.agent', 'entry_point': 'main'},
    'FTSAerotechAgent': {'module': 'socs.agents.fts_aerotech.agent', 'entry_point': 'main'},
    'HWPBBBAgent': {'module': 'socs.agents.hwp_encoder.agent', 'entry_point': 'main'},
    'HWPPicoscopeAgent': {'module': 'socs.agents.hwp_picoscope.agent', 'entry_point': 'main'},
    'ibootbarAgent': {'module': 'socs.agents.ibootbar.agent', 'entry_point': 'main'},
    'LabJackAgent': {'module': 'socs.agents.labjack.agent', 'entry_point': 'main'},
    'Lakeshore240Agent': {'module': 'socs.agents.lakeshore240.agent', 'entry_point': 'main'},
    'Lakeshore336Agent': {'module': 'socs.agents.lakeshore336.agent', 'entry_point': 'main'},
    'Lakeshore370Agent': {'module': 'socs.agents.lakeshore370.agent', 'entry_point': 'main'},
    'Lakeshore372Agent': {'module': 'socs.agents.lakeshore372.agent', 'entry_point': 'main'},
    'Lakeshore425Agent': {'module': 'socs.agents.lakeshore425.agent', 'entry_point': 'main'},
    'LATRtXYStageAgent': {'module': 'socs.agents.xy_stage.agent', 'entry_point': 'main'},
    'MagpieAgent': {'module': 'socs.agents.magpie.agent', 'entry_point': 'main'},
    'MeinbergM1000Agent': {'module': 'socs.agents.meinberg_m1000.agent', 'entry_point': 'main'},
    'PfeifferAgent': {'module': 'socs.agents.pfeiffer_tpg366.agent', 'entry_point': 'main'},
    'PfeifferTC400Agent': {'module': 'socs.agents.pfeiffer_tc400.agent', 'entry_point': 'main'},
    'PysmurfController': {'module': 'socs.agents.pysmurf_controller.agent', 'entry_point': 'main'},
    'PysmurfMonitor': {'module': 'socs.agents.pysmurf_monitor.agent', 'entry_point': 'main'},
    'RotationAgent': {'module': 'socs.agents.hwp_rotation.agent', 'entry_point': 'main'},
    'ScpiPsuAgent': {'module': 'socs.agents.scpi_psu.agent', 'entry_point': 'main'},
    'SmurfFileEmulator': {'module': 'socs.agents.smurf_file_emulator.agent', 'entry_point': 'main'},
    'SmurfStreamSimulator': {'module': 'socs.agents.smurf_stream_simulator.agent', 'entry_point': 'main'},
    'SupRsync': {'module': 'socs.agents.suprsync.agent', 'entry_point': 'main'},
    'SynaccessAgent': {'module': 'socs.agents.synacc.agent', 'entry_point': 'main'},
    'SynthAgent': {'module': 'socs.agents.holo_synth.agent', 'entry_point': 'main'},
    'TektronixAWGAgent': {'module': 'socs.agents.tektronix3021c.agent', 'entry_point': 'main'},
    'ThorlabsMC2000BAgent': {'module': 'socs.agents.thorlabs_mc2000b.agent', 'entry_point': 'main'},
    'VantagePro2Agent': {'module': 'socs.agents.vantagepro2.agent', 'entry_point': 'main'},
    'WiregridActuatorAgent': {'module': 'socs.agents.wiregrid_actuator.agent', 'entry_point': 'main'},
    'WiregridEncoderAgent': {'module': 'socs.agents.wiregrid_encoder.agent', 'entry_point': 'main'},
    'WiregridKikusuiAgent': {'module': 'socs.agents.wiregrid_kikusui.agent', 'entry_point': 'main'},
}



# 4/2009 BAS
#  read() replicates behavior of pyserial
#  readexactly() added, which is probably more useful
#  readbuf() dumps current buffer contents
#  readpacket() has old broken behavior of read() - lowest level / fastest

import socket
import time

MOXA_DEFAULT_TIMEOUT = 1.0

# Socket modes:
# Nonblocking
#  s.setblocking(0) means s.settimeout(0)
#  Read returns as much data as possible, does not fail
#  Doesn't work for me on windows.  Don't use.
# Timeout
#  s.settimeout(n)
#  Waits until buffer has enough data, then returns
#  Throws exception (caught by read) if not enough data is ready after n second.
#  Read returns '' on fail
# Blocking
#  s.setblocking(1) or s.settimeout(None)
#  Waits forever until buffer has enough data, then returns
#  This is the default mode for sockets
#  Check socket.getdefaulttimeout() to see what mode sockets are created in

# pyserial style wrapper over IA 5250 TCP Server mode


class Serial_TCPServer(object):
    """Class to speak with the moxa serial / Ethernet converter.
    Set up the moxa box ports according to the specifications of the device
    hooked into each serial port.

    A typical sequence of messages for dealing with a device. Create the
    socket once::

        >>> moxa = moxa_serial.Serial_TCPServer(('IP',port),timeout=1.0)

    Then do this sequence, complicated in some way by an individual device's hand
    shaking needs::

        >>> moxa.flushInput()
        >>> moxa.write(msg)
        >>> moxa.readexactly(n)

    I write a "cmd" methods that handle the proper sequence with checksums etc.
    Most devices require a certain delay between commands, which is left to the
    user. If using multithreading, wrap your delays in mutexes.

    Args:
        port (tuple): (IP addr, TCP port)
        timeout (float): Timeout for reading from the moxa box

    """

    def __init__(self, port, timeout=MOXA_DEFAULT_TIMEOUT):
        self.port = port

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(0)
        self.settimeout(timeout)
        self.sock.connect(self.port)

    def readexactly(self, n):
        """Tries to read exactly n bytes within the timeout.

        Args:
            n: Number of bytes to read.

        Returns:
            str: Returned message if n bytes were read. Empty string if
            ``len(message) != n``.

        """
        t0 = time.time()
        msg = ""
        timeout = self.gettimeout()
        while len(msg) < n:
            newtimeout = timeout - (time.time() - t0)
            if newtimeout <= 0.0:
                break
            self.settimeout(newtimeout)
            try:
                msg = self.sock.recv(n, socket.MSG_PEEK)
            except BaseException:
                pass
        # Flush the message out if you got everything
        if len(msg) == n:
            msg = self.sock.recv(n).decode()
        # Otherwise tell nothing and leave the data in the buffer
        else:
            msg = ''
        self.settimeout(timeout)
        return msg

    def readbuf_slow(self, n):
        """Reads whatever is in the buffer right now, but is O(N) in buffer
        size.

        Args:
            n: Number of bytes to read.

        """
        msg = ''
        self.sock.setblocking(0)
        try:
            for i in range(n):
                msg += self.sock.recv(1)
        except BaseException:
            pass
        self.sock.setblocking(1)  # belt and suspenders
        self.settimeout(self.__timeout)
        return msg

    def readbuf(self, n):
        """Returns whatever is currently in the buffer. Suitable for large
        buffers.

        Args:
            n: Number of bytes to read.

        """
        if n == 0:
            return ''
        try:
            msg = self.sock.recv(n)
        except BaseException:
            msg = ''
        n2 = min(n - len(msg), n / 2)
        return msg + self.readbuf(n2)

    def readpacket(self, n):
        """Like ``read()``, but may not return everything if the moxa box
        flushes too soon.

        Will probably read whatever arrives in the buffer, up to n or the
        timeout. Use ``read()`` for certainty.

        """
        try:
            msg = self.sock.recv(n)
        except BaseException:
            msg = ''
        return msg

    def read(self, n):
        """Like ``readexactly()``, but returns whatever is in the buffer if it
        can't fill up.

        This replicates the behavior of the read method in pyserial. I feel
        that ``readexactly()`` has better behavior for most applications
        though.

        Args:
            n: Number of bytes to read. Will read at most n bytes.

        Returns:
            str: Returned message of up to n bytes.

        """
        msg = self.readexactly(n)
        n2 = n - len(msg)
        if n2 > 0:
            msg += self.readbuf(n2)
        return msg

    def readline(self, term='\n'):
        msg = ''
        while True:
            c = self.readexactly(1)
            if c == term or c == '':
                return msg
            msg += c

    def readall(self):
        msg = ""
        while True:
            c = self.readexactly(1)
            if c == '\r':
                return msg
            if c == '':
                return False
            msg += c
        return msg

    def write(self, msg):
        """Sends message to the moxa box.

        Args:
            msg (str): Message to send, including terminator (i.e. ``\\r\\n``) if
                needed.

        """
        self.sock.send(msg.encode())

    def writeread(self, msg):
        self.flushInput()
        self.write(msg)
        return self.readall()

    def flushInput(self):
        """Erases the input buffer at this moment.

        Before I ask for new info from a device, I flush my
        receive buffer to make sure I don't get any garbage in
        front.

        """
        self.sock.setblocking(0)
        try:
            while len(self.sock.recv(1)) > 0:
                pass
        except BaseException:
            pass
        self.sock.setblocking(1)
        self.sock.settimeout(self.__timeout)

    def settimeout(self, timeout):
        """Sets the socket in timeout mode."""
        assert timeout > 0.0
        self.__timeout = timeout
        self.sock.settimeout(timeout)
        # We don't query the socket's timeout or check that they're still
        # correct. Since self.sock e is public this could be the wrong
        # timeout!

    def gettimeout(self):
        return self.__timeout

    timeout = property(gettimeout, settimeout,
                       doc='Communication timeout. Only use timeout mode '
                           + 'with ``timeout > 0.0``.')



import argparse
import time

import numpy as np
from ocs import ocs_agent, site_config
from ocs.ocs_client import OCSClient
from ocs.ocs_twisted import TimeoutLock

from socs.agents.wiregrid_kikusui.drivers.common import openlog, writelog
from socs.common import pmx


class WiregridKikusuiAgent:
    """Agent to control the wire-grid rotation
    The KIKUSUI is a power supply and
    it is controlled via serial-to-ethernet converter.
    The converter is linked to the KIKUSUI
    via RS-232 (D-sub 9pin cable).
    The agent communicates with the converter via eternet.

    Args:
        kikusui_ip (str): IP address of the serial-to-ethernet converter
        kikusui_port (int or str): Asigned port for the KIKUSUI power supply
            The converter has four D-sub ports to control
            multiple devices connected via serial communication.
            Communicating device is determined
            by the ethernet port number of the converter.
        encoder_agent (str): Instance ID of the wiregrid encoder agent
        debug (bool): ON/OFF of writing a log file
    """

    def __init__(self, agent, kikusui_ip, kikusui_port,
                 encoder_agent='wgencoder', debug=False):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.take_data = False
        self.kikusui_ip = kikusui_ip
        self.kikusui_port = int(kikusui_port)
        self.encoder_agent = encoder_agent
        self.debug = debug

        self.position_path = '/data/wg-data/position.log'
        self.debug_log_path = '/data/wg-data/action/'

        self.open_trial = 10
        self.Deg = 360 / 52000
        # self.feedback_time = [0.151, 0.241, 0.271, 0.301, 0.331]
        self.feedback_time = [0.151, 0.241, 0.271, 0.361, 0.451]
        self.feedback_cut = [1., 2., 3., 5.0, 8.0]
        self.feedback_steps = 8
        self.num_laps = 2
        self.stopped_time = 10
        self.agent_interval = 0.1

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'kikusui_psu', record=True, agg_params=agg_params)

        try:
            self.PMX = pmx.PMX(
                tcp_ip=self.kikusui_ip,
                tcp_port=self.kikusui_port,
                timeout=0.5)
        except Exception as e:
            self.log.warn(
                'Could not connect to serial converter! | Error = "%s"' % e)
            self.PMX = None

        if self.PMX is not None:
            self.cmd = pmx.Command(self.PMX)
        else:
            self.cmd = None

        # Connect to the encoder agent
        self.encoder_clident = None
        self._connect_encoder()

    ######################
    # Internal functions #
    ######################

    def _check_connect(self):
        if self.PMX is None:
            msg = 'No connection to the KIKUSUI power supply. | '\
                  'Error = "PMX is None"'
            self.log.warn(msg)
            return False, msg
        else:
            msg, ret = self.PMX.check_connect()
            if not ret:
                msg = 'No connection to the KIKUSUI power supply. | '\
                      'Error = "%s"' % msg
                self.log.warn(msg)
                return False, msg
        return True, 'Connection is OK.'

    def _reconnect(self):
        self.log.warn('Trying to reconnect...')
        # reconnect
        try:
            if self.PMX:
                del self.PMX
            if self.cmd:
                del self.cmd
            self.PMX = pmx.PMX(
                tcp_ip=self.kikusui_ip,
                tcp_port=self.kikusui_port,
                timeout=0.5)
        except Exception as e:
            msg = 'Could not reconnect to the KIKUSUI power supply! | '\
                  'Error: %s' % e
            self.log.warn(msg)
            self.PMX = None
            self.cmd = None
            return False, msg
        # reinitialize cmd
        self.cmd = pmx.Command(self.PMX)
        ret, msg = self._check_connect()
        if ret:
            msg = 'Successfully reconnected to the KIKUSUI power supply!'
            self.log.info(msg)
            return True, msg
        else:
            msg = 'Failed to reconnect to the KIKUSUI power supply!'
            self.log.warn(msg)
            if self.PMX:
                del self.PMX
            if self.cmd:
                del self.cmd
            self.PMX = None
            self.cmd = None
            return False, msg

    def _connect_encoder(self):
        self.encoder_client = OCSClient(self.encoder_agent)

    def _rotate_alittle(self, operation_time):
        if operation_time != 0.:
            self.cmd.user_input('ON')
            time.sleep(operation_time)
            self.cmd.user_input('OFF')
            time.sleep(self.agent_interval)
            return True, 'Successfully rotate a little!'
        return True, 'No rotation!'

    def _get_position(self):
        position = -1.
        try:
            response = self.encoder_client.acq.status()
        except Exception as e:
            self.log.warn(
                'Failed to get ENCODER POSITION | '
                '{}'.format(e)
            )
            self.log.warn(
                '    --> Retry to connect to the encoder agent'
            )
            self._connect_encoder()
            try:
                response = self.encoder_client.acq.status()
            except Exception as e:
                self.log.error(
                    'Failed to get encoder position | '
                    '{}'.format(e)
                )
                return -1.

        try:
            position = (float)(
                response.session['data']['fields']['enc_data']['reference_degree'][-1])
        except Exception as e:
            self.log.warn(
                'Failed to get encoder position | '
                '{}'.format(e)
            )

        return position

    def _get_exectime(self, position_difference, feedback_cut, feedback_time):
        if position_difference >= feedback_cut[4]:
            operation_time = feedback_time[4]
        if (feedback_cut[4] > position_difference) &\
           (position_difference >= feedback_cut[3]):
            operation_time = feedback_time[3]
        if (feedback_cut[3] > position_difference) &\
           (position_difference >= feedback_cut[2]):
            operation_time = feedback_time[2]
        if (feedback_cut[2] > position_difference) &\
           (position_difference >= feedback_cut[1]):
            operation_time = feedback_time[1]
        if (feedback_cut[1] > position_difference) &\
           (position_difference >= feedback_cut[0]):
            operation_time = feedback_time[0]
        if feedback_cut[0] > position_difference:
            operation_time = 0.
        return operation_time

    def _move_next(self, logfile, feedback_steps, feedback_time):
        wanted_angle = 22.5
        uncertaity_cancel = 3
        absolute_position = np.arange(0, 360, wanted_angle)

        start_position = self._get_position()
        if (360 < start_position + uncertaity_cancel):
            goal_position = wanted_angle
        elif absolute_position[-1] < start_position + uncertaity_cancel:
            goal_position = 0
        else:
            goal_position = min(
                absolute_position[np.where(
                    start_position + uncertaity_cancel < absolute_position)[0]
                ]
            )

        if self.debug:
            writelog(logfile, 'ON', 0, start_position, 'stepwise')

        self._rotate_alittle(feedback_time[-1] + 0.1)
        time.sleep(self.agent_interval)

        for step in range(feedback_steps):
            mid_position = self._get_position()
            if goal_position + wanted_angle < mid_position:
                operation_time =\
                    self._get_exectime(
                        goal_position - (mid_position - 360),
                        self.feedback_cut,
                        feedback_time)
            else:
                operation_time =\
                    self._get_exectime(
                        goal_position - mid_position,
                        self.feedback_cut,
                        feedback_time)

            if operation_time == 0.:
                break
            self._rotate_alittle(operation_time)

        if self.debug:
            writelog(logfile, 'OFF', 0,
                     self._get_position(), 'stepwise')

    ##################
    # Main functions #
    ##################

    def set_on(self, session, params=None):
        """set_on()

        **Task** - Set output ON.

        """
        with self.lock.acquire_timeout(timeout=5, job='set_on') as acquired:
            if not acquired:
                self.log.warn('Could not set ON because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock'

            if self.debug:
                logfile = openlog(self.debug_log_path)
                writelog(logfile, 'ON', 0,
                         self._get_position(), 'continuous')
                logfile.close()

            self.cmd.user_input('ON')
            return True, 'Set Kikusui on'

    def set_off(self, session, params=None):
        """set_off()

        **Task** - Set output OFF.

        """
        with self.lock.acquire_timeout(timeout=5, job='set_off') as acquired:
            if not acquired:
                self.log.warn('Could not set OFF because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock'

            if self.debug:
                logfile = openlog(self.debug_log_path)
                writelog(logfile, 'OFF', 0,
                         self._get_position(), 'continuous')
                logfile.close()

            self.cmd.user_input('OFF')
            return True, 'Set Kikusui off'

    @ocs_agent.param('current', default=0., type=float,
                     check=lambda x: 0.0 <= x <= 3.0)
    def set_c(self, session, params):
        """set_c(current=0)

        **Task** - Set current [A]

        Parameters:
            current (float): set current [A] (should be [0.0, 3.0])
        """
        current = params.get('current')

        with self.lock.acquire_timeout(timeout=5, job='set_c') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not run set_c() because {} is already running'
                    .format(self.lock.job))
                return False, 'Could not acquire lock'

            self.cmd.user_input('C {}'.format(current))
            return True, 'Set Kikusui current to {} A'.format(current)

    @ocs_agent.param('volt', default=12., type=float,
                     check=lambda x: 0.0 <= x <= 12.0)
    def set_v(self, session, params):
        """set_v*volt=12.)

        **Task** - Set voltage [V].

        Parameters:
            volt: set voltage [V] (Usually 12V / Should be [0.0, 12.0])
        """
        volt = params.get('volt')

        with self.lock.acquire_timeout(timeout=5, job='set_v') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not run set_v() because {} is already running'
                    .format(self.lock.job))
                return False, 'Could not acquire lock'

            self.cmd.user_input('V {}'.format(volt))

            return True, 'Set Kikusui voltage to {} V'.format(volt)

    def get_vc(self, session, params=None):
        """get_vc()

        **Task** - Show voltage [V], current [A], output on/off.

        """
        with self.lock.acquire_timeout(timeout=5, job='get_vc') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not run get_vc() because {} is already running'
                    .format(self.lock.job))
                return False, 'Could not acquire lock'

            v_val = None
            c_val = None
            s_val = None
            msg = 'Error'
            s_msg = 'Error'

            # check connection
            ret, msg = self._check_connect()
            if not ret:
                msg = 'Could not get c,v because of failure of connection.'
                self.log.warn(msg)
            else:
                v_val, c_val = self.cmd.user_input('VC?')
                s_msg, s_val = self.cmd.user_input('O?')

            return True,\
                'Get Kikusui voltage / current: {} V / {} A [status={}]'\
                .format(v_val, c_val, s_val)

    def get_angle(self, session, params=None):
        """get_angle()

        **Task** - Show wire-grid rotaiton angle [deg].

        """
        with self.lock.acquire_timeout(timeout=5, job='get_angle') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not run get_angle() because {} is already running'
                    .format(self.lock.job))
                return False, 'Could not acquire lock'

            angle = self._get_position()
            if angle < 0.:
                return False,\
                    'Could not get the angle of the wire-grid rotation.'

            return True,\
                'Get wire-grid rotation angle = {} deg'.format(angle)

    @ocs_agent.param('storepath', default='/data/wg-data/action/', type=str)
    def calibrate_wg(self, session, params):
        """calibrate_wg(storepath='/data/wg-data/action/')

        **Task** - Run rotation-motor calibration for wire-grid.

        Parameters:
            storepath (str): Path for log file.
        """
        storepath = params.get('storepath')

        with self.lock.acquire_timeout(timeout=5, job='calibrate_wg')\
                as acquired:
            if not acquired:
                self.log.warn('Could not run calibrate_wg() '
                              'because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock'

            logfile = openlog(storepath)

            cycle = 1
            for i in range(11):
                tperiod = 0.10 + 0.02 * i
                for j in range(100):
                    if j % 20 == 0:
                        self.log.warn(f'this is {cycle}th time action')

                    writelog(logfile, 'ON', tperiod,
                             self._get_position(), 'calibration')
                    self._rotate_alittle(tperiod)
                    time.sleep(self.agent_interval + 1.)
                    writelog(logfile, 'OFF', 0.,
                             self._get_position(), 'calibration')
                    writelog(logfile, 'ON', 0.70,
                             self._get_position(), 'calibration')
                    self._rotate_alittle(0.70)
                    time.sleep(self.agent_interval + 1.)
                    writelog(logfile, 'OFF', 0.,
                             self._get_position(), 'calibration')
                    cycle += 1

            logfile.close()

            return True,\
                'Micro step rotation of wire grid finished. '\
                'Please calibrate and take feedback params.'

    def stepwise_rotation(self, session, params=None):
        """stepwise_rotation(feedback_steps=8, num_laps=1, stopped_time=10, \
                             feedback_time=[0.181, 0.221, 0.251, 0.281, 0.301])

        **Task** - Run step-wise rotation for wire-grid calibration. In each
        step, seveal small-rotations are performed to rotate 22.5-deg.

        Parameters:
            feedback_steps (int): Number of small rotations
                                  for each 22.5-deg step.
            num_laps (int): Number of laps (revolutions).
            stopped_time (float): Stopped time [sec]
                                  for each 22.5-deg step.
            feedback_time (list): Calibration constants
                                  for the 22.5-deg rotation.
        """
        if params is None:
            params = {}

        with self.lock.acquire_timeout(timeout=5, job='stepwise_rotation')\
                as acquired:
            if not acquired:
                self.log.warn('Could not run stepwise_rotation() '
                              'because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock'

            self.feedback_steps = params.get('feedback_steps', 8)
            self.num_laps = params.get('num_laps', 1)
            self.stopped_time = params.get('stopped_time', 10)
            self.feedback_time = params.get(
                'feedback_time', [0.181, 0.221, 0.251, 0.281, 0.301])

            logfile = openlog(self.debug_log_path)

            for i in range(int(self.num_laps * 16.)):
                self._move_next(
                    logfile, self.feedback_steps, self.feedback_time)
                time.sleep(self.stopped_time)

            logfile.close()

            return True, 'Step-wise rotation finished'

    def IV_acq(self, session, params=None):
        """IV_acq()

        **Process** - Run data acquisition.

        Notes:
            The most recent data collected is stored in session.data in the
            structure::

                >>> response.session['data']
                {'fields':
                    {
                     'kikusui':
                        {'volt': voltage [V],
                         'curr': current [A],
                         'voltset': voltage setting [V],
                         'currset': current setting [A],
                         'status': output power status 1(on) or 0(off)
                         }
                    },
                 'timestamp':1601925677.6914878
                }

        """

        # timeout is long because calibrate_wg will take a long time (> hours)
        with self.lock.acquire_timeout(timeout=1000, job='IV_acq') as acquired:
            if not acquired:
                self.log.warn('Could not run IV_acq '
                              'because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock'

            session.set_status('running')

            self.take_data = True
            last_release = time.time()
            session.data = {'fields': {}}
            while self.take_data:
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=1000):
                        self.log.warn(
                            'IV_acq(): '
                            'Could not re-acquire lock now held by {}.'
                            .format(self.lock.job))
                        if self.lock.job == 'calibrate_wg':
                            self.log.warn(
                                'IV_acq(): '
                                'Continue to wait for {}()'
                                .format(self.lock.job))
                            # Wait for lock acquisition forever
                            acquired = self.lock.acquire(
                                timeout=-1, job='IV_acq')
                            if acquired:
                                self.log.info(
                                    'Succcessfully got the lock '
                                    'for IV_acq()!')
                            else:
                                self.log.warn(
                                    'Failed to aquire the lock '
                                    'for IV_acq()!')
                                return False,\
                                    'Could not re-acquire lock '\
                                    'for IV_acq() (timeout=-1)'
                        else:
                            return False,\
                                'Could not re-acquire lock '\
                                'for IV_acq() (timeout=1000 sec)'

                current_time = time.time()
                data = {'timestamp': time.time(),
                        'block_name': 'Kikusui_IV',
                        'data': {}}

                # check connection
                ret, msg = self._check_connect()
                if not ret:
                    msg = 'Could not connect to the KIKUSUI power supply!'
                    v_val, i_val, vs_val, is_val = 0., 0., 0., 0.
                    s_val = -1  # -1 means Not connected.
                    self.log.warn(msg)
                    # try to reconnect
                    self._reconnect()
                else:
                    v_msg, v_val = self.cmd.user_input('V?')
                    i_msg, i_val = self.cmd.user_input('C?')
                    vs_msg, vs_val = self.cmd.user_input('VS?')
                    is_msg, is_val = self.cmd.user_input('CS?')
                    s_msg, s_val = self.cmd.user_input('O?')
                data['data']['kikusui_volt'] = v_val
                data['data']['kikusui_curr'] = i_val
                data['data']['kikusui_voltset'] = vs_val
                data['data']['kikusui_currset'] = is_val
                data['data']['kikusui_status'] = s_val
                self.agent.publish_to_feed('kikusui_psu', data)
                # store session.data
                field_dict = {'kikusui':
                              {'volt': v_val,
                               'curr': i_val,
                               'voltset': vs_val,
                               'currset': is_val,
                               'status': s_val
                               }
                              }
                session.data['fields'] = field_dict
                session.data['timestamp'] = current_time

                time.sleep(1)  # DAQ interval
            # End of while loop for take_data
        # End of acquired lock

        self.agent.feeds['kikusui_feed'].flush_buffer()
        return True, 'Acqusition exited cleanly'

    def stop_IV_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data'

        return False, 'acq is not currently running'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--kikusui-ip')
    pgroup.add_argument('--kikusui-port')
    pgroup.add_argument('--encoder-agent', dest='encoder_agent',
                        default='wgencoder',
                        help='Instance id of the wiregrid encoder agent')
    pgroup.add_argument('--debug', dest='debug',
                        action='store_true', default=False,
                        help='Write a log file for debug')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='WiregridKikusuiAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    kikusui_agent = WiregridKikusuiAgent(agent, kikusui_ip=args.kikusui_ip,
                                         kikusui_port=args.kikusui_port,
                                         encoder_agent=args.encoder_agent,
                                         debug=args.debug)
    agent.register_process('IV_acq', kikusui_agent.IV_acq,
                           kikusui_agent.stop_IV_acq, startup=True)
    agent.register_task('set_on', kikusui_agent.set_on)
    agent.register_task('set_off', kikusui_agent.set_off)
    agent.register_task('set_c', kikusui_agent.set_c)
    agent.register_task('set_v', kikusui_agent.set_v)
    agent.register_task('get_vc', kikusui_agent.get_vc)
    agent.register_task('get_angle', kikusui_agent.get_angle)
    agent.register_task('calibrate_wg', kikusui_agent.calibrate_wg)
    agent.register_task('stepwise_rotation',
                        kikusui_agent.stepwise_rotation)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()
